"""Segment display on the FAST platform."""

from typing import List

from mpf.core.utility_functions import Util
from mpf.core.rgb_color import RGBColor
from mpf.devices.segment_display.segment_display_text import ColoredSegmentDisplayText

from mpf.platforms.interfaces.segment_display_platform_interface \
    import SegmentDisplayPlatformInterface, FlashingType


class FASTSegmentDisplay(SegmentDisplayPlatformInterface):

    """FAST segment display."""

    __slots__ = ["serial", "hex_id", "next_color", "next_text", "current_color"]

    def __init__(self, index, communicator):
        """Initialize alpha numeric display."""
        super().__init__(index)
        self.serial = communicator
        self.hex_id = Util.int_to_hex_string(index * 7)
        self.next_color = None
        self.next_text = None
        self.current_color = None

    def set_text(self, text: ColoredSegmentDisplayText, flashing: FlashingType, flash_mask: str) -> None:
        """Set digits to display."""
        del flashing
        del flash_mask
        colors = text.get_colors()
        self.next_text = text
        if colors:
            self._set_color(colors)

    def _set_color(self, colors: List[RGBColor]) -> None:
        """Set display color."""
        if len(colors) == 1:
            next_color = (RGBColor(colors[0]).hex + ',') * 7
        else:
            next_color = ','.join([RGBColor(color).hex for color in colors]) + ','

        # Current color is set by the FastSegCommunicator after the serial
        # command is written to set the color (which happens automatically
        # on the next loop whenever self.next_color is set).
        if next_color != self.current_color:
            self.next_color = next_color



"""Service mode for MPF."""
# import subprocess
# import os
from collections import namedtuple

from typing import List

from mpf.core.async_mode import AsyncMode
from mpf.core.switch_controller import MonitoredSwitchChange
from mpf.core.utility_functions import Util

ServiceMenuEntry = namedtuple("ServiceMenuEntry", ["label", "callback"])
LightChainMap = namedtuple("LightMap", ["board", "chain", "light"])


class Service(AsyncMode):

    """The service mode."""

    __slots__ = ("_do_sort", "_is_displayed", "_menu_level", "_trigger",  "_update_script")

    def __init__(self, *args, **kwargs):
        """Initialize service mode."""
        super().__init__(*args, **kwargs)
        self._update_script = None
        self._do_sort = self.config.get('mode_settings', {}).get('sort_devices_by_number', True)
        self._menu_level = -1
        self._trigger = None
        self._is_displayed = False

    def mode_start(self, **kwargs):
        """Create an event handler for the "reset" event triggered via keypress."""
        del kwargs
        self.add_mode_event_handler("reset", self._on_reset)

        # Map MC-triggered events to methods in the parent class
        self.add_mode_event_handler("service_trigger", self._on_service_trigger)

    @staticmethod
    def get_config_spec():
        """Add validation for mode."""
        return '''
door_opened_events: list|str|service_door_opened
door_closed_events: list|str|service_door_closed
high_power_on_events: list|str|service_power_on
high_power_off_events: list|str|service_power_off
enter_events: list|str|sw_service_enter_active
esc_events: list|str|sw_service_esc_active
up_events: list|str|sw_service_up_active
down_events: list|str|sw_service_down_active
software_update: single|bool|False
software_update_script: single|str|None
sort_devices_by_number: single|bool|True
'''

    async def _service_mode_exit(self):
        await self.machine.service.stop_service()

    async def _get_key(self):
        futures = {
            self.machine.events.wait_for_any_event(
                ["sw_service_down_active"]
            ): "DOWN",
            self.machine.events.wait_for_any_event(
                ["sw_service_up_active"]
            ): "UP",
            self.machine.events.wait_for_any_event(
                ["sw_service_toggle_active", "sw_service_toggle_inactive"]
            ): "TOGGLE",
            self.machine.events.wait_for_any_event(
                ["sw_service_enter_active"]
            ): "ENTER",
            self.machine.events.wait_for_any_event(
                ["sw_service_esc_active"]
            ): "ESC",
            self.machine.events.wait_for_any_event(
                ["s_flipper_left_inactive"]
            ): "PAGE_LEFT",
            self.machine.events.wait_for_any_event(
                ["s_flipper_right_inactive"]
            ): "PAGE_RIGHT",
            self.machine.events.wait_for_any_event(
                # Use the INACTIVE event to prevent attract from starting a game
                # (which it does on inactive)
                ["sw_start_inactive"]
            ): "START",
            self.machine.events.wait_for_any_event(
                ["service_trigger"]
            ): "TRIGGER"
        }

        key = await Util.race(futures)
        if key == "TRIGGER":
            await self._service_trigger()
            key = None
        return key

    async def _run(self):
        while True:
            # wait for key
            key = await self._get_key()

            if key == "ENTER":
                # start main menu
                await self._start_main_menu()
            elif key == "UP":
                volume = self.machine.variables.get_machine_var("master_volume")
                if not isinstance(volume, float):
                    volume = .5
                volume += .05
                if volume >= 1.0:
                    volume = 1.0
                else:
                    volume = round(volume, 2)
                self.machine.variables.set_machine_var("master_volume", volume)
                # post event for increased volume
                self.machine.events.post("master_volume_increase", volume=volume)
                '''event: master_volume_increase

                desc: Increase the master volume of the audio system.

                args:

                volume: New volume as float between 0.0 an 1.0
                '''
            elif key == "DOWN":
                volume = self.machine.variables.get_machine_var("master_volume")
                if not isinstance(volume, float):
                    volume = .5
                volume -= .05
                if volume <= 0.0:
                    volume = 0.0
                else:
                    volume = round(volume, 2)
                self.machine.variables.set_machine_var("master_volume", volume)
                # post event for decreased volume
                self.machine.events.post("master_volume_decrease", volume=volume)
                '''event: master_volume_decrease

                desc: Decrease the master volume of the audio system.

                args:

                volume: New volume as float between 0.0 an 1.0
                '''

    async def _start_main_menu(self):
        self.machine.service.start_service()
        self.machine.events.post("service_main_menu")
        await self._service_mode_main_menu()

        await self._service_mode_exit()

    async def _service_mode_main_menu(self):
        self._is_displayed = True
        while self._is_displayed:
            key = await self._get_key()
            if key:
                self.machine.events.post("service_button", button=key)

    # # Utilities
    # def _load_utilities_menu_entries(self) -> List[ServiceMenuEntry]:
    #     """Return the utilities menu items with label and callback."""
    #     entries = [
    #         ServiceMenuEntry("Reset Menu", self._utilities_reset_menu),
    #     ]

    #     if self.config['mode_settings']['software_update']:
    #         update_file_path = self.config['mode_settings']['software_update_script']
    #         if not update_file_path:
    #             raise AssertionError("Please configure software_update_script to enable software_update in "
    #                                  "service mode.")

    #         if not os.path.isabs(update_file_path):
    #             update_file_path = os.path.join(self.machine.machine_path, update_file_path)

    #         if os.path.isfile(update_file_path):
    #             self._update_script = update_file_path
    #             entries.append(ServiceMenuEntry("Software Update", self._software_update))

    #     return entries

    # async def _utilities_menu(self):
    #     await self._make_menu(self._load_utilities_menu_entries())

    # def _load_utilities_reset_menu_entries(self) -> List[ServiceMenuEntry]:
    #     """Return the utilities reset menu items with label and callback."""
    #     return [
    #         ServiceMenuEntry("Reset Coin Audits", self._utilities_reset_coin_audits),
    #         ServiceMenuEntry("Reset Game Audits", self._utilities_reset_game_audits),
    #         ServiceMenuEntry("Reset High Scores", self._utilities_reset_high_scores),
    #         ServiceMenuEntry("Reset Credits", self._utilities_reset_credits),
    #         ServiceMenuEntry("Reset to Factory Settings", self._utilities_reset_to_factory_settings),
    #     ]

    # async def _utilities_reset_menu(self):
    #     await self._make_menu(self._load_utilities_reset_menu_entries())

    # async def _utilities_reset_coin_audits(self):
    #     selection = await self._make_option_slide("Reset earning audits", "Perform coin reset?", ["no", "yes"],
    #                                               "THIS CANNOT BE UNDONE")
    #     if selection == "yes":
    #         self.machine.events.post("earnings_reset")

    # async def _utilities_reset_game_audits(self):
    #     selection = await self._make_option_slide("Auditor Reset", "Reset Game Audits?", ["no", "yes"],
    #                                               "THIS CANNOT BE UNDONE")
    #     if selection == "yes":
    #         self.machine.events.post("auditor_reset")

    # async def _utilities_reset_high_scores(self):
    #     selection = await self._make_option_slide("High Score Reset", "Remove Highscores?", ["no", "yes"],
    #                                               "THIS CANNOT BE UNDONE")
    #     if selection == "yes":
    #         self.machine.events.post("high_scores_reset")

    # async def _utilities_reset_credits(self):
    #     selection = await self._make_option_slide("Reset credits", "Remove all credits?", ["no", "yes"],
    #                                               "THIS CANNOT BE UNDONE")
    #     if selection == "yes":
    #         self.machine.events.post("credits_reset")

    # async def _utilities_reset_to_factory_settings(self):
    #     selection = await self._make_option_slide("Factory Reset", "Reset to factory setting?", ["no", "yes"],
    #                                               "THIS CANNOT BE UNDONE")
    #     if selection == "yes":
    #         self.machine.events.post("factory_reset")

    # async def _software_update(self):
    #     run_update = False
    #     self._update_software_update_slide(run_update)

    #     while True:
    #         key = await self._get_key()
    #         if key == 'ESC':
    #             break
    #         if key in ('UP', 'DOWN'):
    #             run_update = not run_update
    #             self._update_software_update_slide(run_update)
    #         elif key == 'ENTER' and run_update:
    #             # perform update
    #             self.machine.events.post("service_software_update_start")
    #             # pylint: disable-msg=consider-using-with
    #             subprocess.Popen([self._update_script])
    #             self.machine.stop("Software Update")

    #     self.machine.events.post("service_software_update_stop")

    # def _update_software_update_slide(self, run_update):
    #     self.machine.events.post("service_software_update_choice", run_update="Yes" if run_update else "No")

    def _switch_monitor(self, change: MonitoredSwitchChange):
        if change.state:
            state_string = "active"
            self.machine.events.post("service_switch_hit")
        else:
            state_string = "inactive"

        label_string = "" if change.label == "%" else change.label

        self.machine.events.post("service_switch_test_start",
                                 switch_name=change.name,
                                 switch_num=change.num,
                                 switch_label=label_string,
                                 switch_state=state_string)

    async def _switch_test_menu(self):
        self.machine.switch_controller.add_monitor(self._switch_monitor)
        self.machine.events.post("service_switch_test_start",
                                 switch_name="", switch_state="", switch_num="", switch_label="")
        await self.machine.events.wait_for_any_event(self.config['mode_settings']['esc_events'])
        self.machine.events.post("service_switch_test_stop")
        self.machine.switch_controller.remove_monitor(self._switch_monitor)

    def _update_coil_slide(self, items, position):
        board, coil = items[position]
        self.machine.events.post("service_coil_test_start",
                                 board_name=board,
                                 coil_name=coil.name,
                                 coil_label=coil.config['label'],
                                 coil_num=coil.hw_driver.number)

    async def _coil_test_menu(self):
        position = 0
        items = self.machine.service.get_coil_map(do_sort=self._do_sort)

        # do not crash if no coils are configured
        if not items:   # pragma: no cover
            return

        self._update_coil_slide(items, position)

        while True:
            key = await self._get_key()
            if key == 'ESC':
                break
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
                self._update_coil_slide(items, position)
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
                self._update_coil_slide(items, position)
            elif key == 'ENTER':
                # pulse coil
                items[position].coil.pulse()

        self.machine.events.post("service_coil_test_stop")

    def _update_light_slide(self, items, position, color):
        board, light = items[position]
        label_string = "" if light.config['label'] == "%" else light.config['label']
        self.machine.events.post("service_light_test_start",
                                 board_name=board,
                                 light_name=light.name,
                                 light_label=label_string,
                                 light_num=light.config['number'],
                                 test_color=color)

    def _update_light_chain_slide(self, items, position, color):
        board, chain, _ = items[position]  # Unused variable "lights"
        self.machine.events.post("service_light_test_start",
                                 board_name=board,
                                 light_name=" ",
                                 light_label=chain,
                                 light_num=" ",
                                 test_color=color)

    async def _light_test_menu(self):
        position = 0
        color_position = 0
        colors = ["white", "red", "green", "blue", "yellow"]
        items = self.machine.service.get_light_map(do_sort=self._do_sort)

        # do not crash if no lights are configured
        if not items:   # pragma: no cover
            return

        self._update_light_slide(items, position, colors[color_position])

        while True:
            self._update_light_slide(items, position, colors[color_position])
            items[position].light.color(colors[color_position], key="service", priority=1000000)

            key = await self._get_key()
            items[position].light.remove_from_stack_by_key("service")
            if key == 'ESC':
                break
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
            elif key == 'ENTER':
                # change color
                color_position += 1
                if color_position >= len(colors):
                    color_position = 0

        self.machine.events.post("service_light_test_stop")

    async def _light_chain_menu(self):
        position = 0
        color_position = 0
        colors = ["white", "red", "green", "blue", "yellow"]
        items = self._generate_light_chains()

        while True:
            self._update_light_chain_slide(items, position, colors[color_position])
            for _, l in items[position].light:  # Unused variable "addr"
                l.color(colors[color_position], key="service", priority=1000000)

            key = await self._get_key()
            for _, l in items[position].light:  # Unused variable "addr"
                l.remove_from_stack_by_key("service")
            if key == 'ESC':
                break
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
            elif key == 'ENTER':
                # change color
                color_position += 1
                if color_position >= len(colors):
                    color_position = 0

        self.machine.events.post("service_light_test_stop")

    def _generate_light_chains(self):  # pylint: disable=too-many-locals
        items = self.machine.service.get_light_map(do_sort=self._do_sort)

        # Categorize by platform and address
        chain_lookup = {}
        for board, l in items:
            numbers = l.get_hw_numbers()
            chain_2 = None
            # Just choose the first one as representative?
            number = numbers[0]
            if "-" in number:
                bits = number.split("-")  # e.g. led-7-4-r
                if len(bits) == 2:
                    # FAST lights are single addresses in blocks of 64
                    if board.startswith("FAST"):
                        addr = int(bits[0], 16)
                        chain = addr // 64
                    else:
                        chain, addr = bits
                elif len(bits) == 3:
                    chain, addr, _ = bits  # Unused variable "color"
                elif len(bits) == 4:
                    _, chain, addr, _ = bits
                else:
                    self.warning_log("Unknown bits in parsing light address: %s", bits)
                    continue
                chain = f"Chain {chain}"
            elif l.config['subtype'] == "matrix":
                # Matrix lights get two chains: one for the row, one for the column
                number = int(number, 16)
                chain = f"Row {(number // 8) + 1}"
                addr = number % 8
                chain_2 = f"Column {(number % 8) + 1}"
                addr_2 = number // 8
            else:
                chain = "XX"
                addr = number

            for platform in l.platforms:
                platform_name = type(platform).__name__
                if platform_name not in chain_lookup:
                    chain_lookup[platform_name] = {}
                if chain not in chain_lookup[platform_name]:
                    chain_lookup[platform_name][chain] = []
                chain_lookup[platform_name][chain].append((addr, l))
                # This is ugly, but is iteration overkill?
                if chain_2:
                    if chain_2 not in chain_lookup[platform_name]:
                        chain_lookup[platform_name][chain_2] = []
                    chain_lookup[platform_name][chain_2].append((addr_2, l))

        items = []
        for platform_name, chains in chain_lookup.items():
            for chain_name, chain in chains.items():
                items.append(LightChainMap(platform_name, chain_name, chain))
        # do not crash if no lights are configured
        if not items:   # pragma: no cover
            return

        items.sort(key=lambda x: x.chain)

    async def _volume_menu(self, platform=None):
        position = 0
        if platform:
            item_configs = platform.audio_interface.amps
        else:
            item_configs = self.machine.config["sound_system"]["tracks"]
        items = [{
            # TODO: Give each software track a 'name' property
            **config,
            "name": config.get("name", track),
            "label": config.get("label", track),
            "is_platform": bool(platform),
            "value": self.machine.variables.get_machine_var(
                f"{config['name'] if platform else track}_volume") or config['volume']
        } for track, config in item_configs.items()]

        # do not crash if no items
        if not items:   # pragma: no cover
            return

        # Convert floats to ints for systems that use 0.0-1.0 for volume
        for item in items:
            if isinstance(item['value'], float):
                item['value'] = int(item['value'] * 100)

        # If supported on hardware platform, add option to write to firmware
        if platform and hasattr(platform.audio_interface, "save_settings_to_firmware"):
            items.append({
                "name": "write_to_firmware",
                "label": "Write Settings",
                "is_platform": True,
                "value": "Confirm",
                "levels_list": ["Confirm", "Saved"]
            })

        self._update_volume_slide(items, position)

        while True:
            key = await self._get_key()
            if key == 'ESC':
                break
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
                self._update_volume_slide(items, position)
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
                self._update_volume_slide(items, position)
            elif key == 'ENTER':
                # change setting
                await self._volume_change(items, position, platform, focus_change="enter")

        self.machine.events.post("service_volume_stop")

    def _update_volume_slide(self, items, position, is_change=False, focus_change=None):
        config = items[position]
        event = "service_volume_{}".format("edit" if is_change else "start")
        # The 'focus_change' argument can be used to start/stop sound files playing
        # during the service menu, to test volume.
        self.machine.events.post(event,
                                 settings_label=config["label"],
                                 value_label=config["value"],
                                 track=config["name"],
                                 is_platform=config["is_platform"],
                                 focus_change=focus_change)

    async def _volume_change(self, items, position, platform, focus_change=None):
        self._update_volume_slide(items, position, focus_change=focus_change)
        if items[position].get("levels_list"):
            values = items[position]["levels_list"]
        else:
            # Use ints for values to avoid floating-point comparisons
            values = [int((0.05 * i) * 100) for i in range(0, 21)]
        value_position = values.index(items[position]["value"])
        self._update_volume_slide(items, position, is_change=True)

        while True:
            key = await self._get_key()
            new_value = None
            if key == 'ESC':
                self._update_volume_slide(items, position, focus_change="exit")
                break
            if key == 'UP':
                value_position += 1
                if value_position >= len(values):
                    value_position = 0
                new_value = values[value_position]
            elif key == 'DOWN':
                value_position -= 1
                if value_position < 0:
                    value_position = len(values) - 1
                new_value = values[value_position]
            if new_value is not None:
                items[position]['value'] = new_value
                # Check for a firmware update
                if items[position]['name'] == "write_to_firmware":
                    if new_value == "Saved":
                        platform.audio_interface.save_settings_to_firmware()
                        # Remove the options from the list
                        values = ['Saved']
                        items[position]['levels_list'] = values
                else:
                    # Internally tracked values divide by 100 to store a float.
                    # External (hardware) values, use the value units provided
                    # TODO: Create an Amp/Track class to internalize this method.
                    if not items[position].get("levels_list"):
                        new_value = new_value / 100
                    self.machine.variables.set_machine_var(f"{items[position]['name']}_volume", new_value, persist=True)
                self._update_volume_slide(items, position, is_change=True)

    # AUDIT Menu
    def _load_audit_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the audit menu items with label and callback."""
        items = [
            ServiceMenuEntry("Earning Audits", self._audit_earning_menu),
            ServiceMenuEntry("Switch Audits", self._audit_switch_menu),
            ServiceMenuEntry("Shot Audits", self._audit_shot_menu),
            ServiceMenuEntry("Event Audits", self._audit_event_menu),
            ServiceMenuEntry("Player Audits", self._audit_player_menu),
        ]

        if self.machine.auditor.report_missing_switches():
            items.insert(0, ServiceMenuEntry("Missing Switches", self._audit_missing_menu))

        return items

    async def _audits_menu(self):
        await self._make_menu(self._load_audit_menu_entries())

    def _update_audits_slide(self, items, position):
        item = items[position][0]
        value = items[position][1]
        self.machine.events.post("service_audits_menu_show", audits_label=str(item), value_label=str(value))

    async def _audits_submenu(self, items):
        position = 0
        if not items:   # pragma: no cover
            return
        self._update_audits_slide(items, position)
        while True:
            key = await self._get_key()
            if key == 'ESC':
                break
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
                self._update_audits_slide(items, position)
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
                self._update_audits_slide(items, position)
            if key == 'ENTER':
                # fallthrough
                pass
        self.machine.events.post("service_audits_menu_hide")

    async def _audit_earning_menu(self):
        try:
            items = self.machine.modes['credits'].earnings
        except (IndexError, KeyError):
            items = {}

        await self._audits_submenu(list(items.items()))

    async def _audit_missing_menu(self):
        items = self.machine.auditor.report_missing_switches()
        await self._audits_submenu(items)

    async def _audit_player_menu(self):
        audits = self.machine.auditor.current_audits.get('player', {})

        items = []
        for audit, audit_values in audits.items():
            for key, value in audit_values.items():
                items.append(("{} {}".format(audit, key), value))

        await self._audits_submenu(items)

    async def _audit_switch_menu(self):
        items = self.machine.auditor.current_audits.get('switches', {})
        await self._audits_submenu(list(items.items()))

    async def _audit_shot_menu(self):
        items = self.machine.auditor.current_audits.get('shots', {})
        await self._audits_submenu(list(items.items()))

    async def _audit_event_menu(self):
        items = self.machine.auditor.current_audits.get('events', {})
        await self._audits_submenu(list(items.items()))

    def _update_settings_slide(self, items, position, is_change=False):
        setting = items[position]
        label = self.machine.settings.get_setting_value_label(setting.name)
        event = "service_settings_{}".format("edit" if is_change else "start")
        self.machine.events.post(event,
                                 settings_label=setting.label,
                                 value_label=label)

    async def _settings_menu(self, settings_type):
        position = 0
        items = self.machine.settings.get_settings(settings_type)

        # do not crash if no settings
        if not items:   # pragma: no cover
            return

        self._update_settings_slide(items, position)

        while True:
            key = await self._get_key()
            if key == 'ESC':
                break
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
                self._update_settings_slide(items, position)
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
                self._update_settings_slide(items, position)
            elif key == 'ENTER':
                # change setting
                await self._settings_change(items, position)

        self.machine.events.post("service_settings_stop")

    async def _settings_change(self, items, position):
        self._update_settings_slide(items, position)

        values = list(items[position].values.keys())
        value_position = values.index(self.machine.settings.get_setting_value(items[position].name))
        self._update_settings_slide(items, position, is_change=True)

        while True:
            key = await self._get_key()
            if key == 'ESC':
                self._update_settings_slide(items, position)
                break
            if key == 'UP':
                value_position += 1
                if value_position >= len(values):
                    value_position = 0
                self.machine.settings.set_setting_value(items[position].name, values[value_position])
                self._update_settings_slide(items, position, is_change=True)
            elif key == 'DOWN':
                value_position -= 1
                if value_position < 0:
                    value_position = len(values) - 1
                self.machine.settings.set_setting_value(items[position].name, values[value_position])
                self._update_settings_slide(items, position, is_change=True)

## Added for MPF 0.80
##-------------------

    def _on_reset(self, **kwargs):
        del kwargs
        for mode in self.machine.modes.values():
            if not mode.active or mode.name == "game":
                continue
            mode.stop()

        # explicitly stop game last
        if self.machine.modes["game"].active:
            self.machine.modes["game"].stop()

        self.machine.clock.loop.create_task(self.machine.reset())

    def _on_service_trigger(self, **kwargs):
        """Callback from an MC to handle a service menu action.

        May have action 'setting' to save a machine setting value,
        or any action defined in _service_trigger().
        """
        action = kwargs.get("action")
        # For settings, write them directly
        if action == "setting":
            name = kwargs.get("variable")
            # Key type is not stored in the SettingEntry, look it up from config
            # Some are manually added, so default to int
            setting_type = self.machine.settings.config[name]['key_type'] \
                if name in self.machine.settings.config else "int"
            value = kwargs.get("value")
            value = float(value) if setting_type == "float" else int(value)
            self.machine.settings.set_setting_value(name, value)
        # Store this action for when the main loop comes around
        else:
            self._trigger = action

    async def _service_trigger(self):
        if self._trigger == "switch_test":
            await self._switch_test_menu()
        elif self._trigger == "coil_test":
            await self._coil_test_menu()
        elif self._trigger == "light_test":
            await self._light_test_menu()
        elif self._trigger == "light_chain_test":
            await self._light_chain_menu()
        elif self._trigger == "service_exit":
            self._menu_level = -1
            # Set display to false will end the main_menu loop and
            # the native service menu will take care of the rest.
            self._is_displayed = False
        self._trigger = None



"""A generic system11 driver overlay.

This is based on the Snux platform to generically support all kinds of System11 platforms.
"""
from typing import Any, Optional, Set, Tuple, Dict

from mpf.core.machine import MachineController
from mpf.core.platform import DriverPlatform, DriverConfig, SwitchSettings, DriverSettings, RepulseSettings, \
    SwitchPlatform, SwitchConfig

from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings

from mpf.core.delays import DelayManager

MYPY = False
if MYPY:   # pragma: no cover
    class SwitchDriverPlatform(DriverPlatform, SwitchPlatform):     # noqa
        pass


# pylint: disable-msg=too-many-instance-attributes
class System11OverlayPlatform(DriverPlatform, SwitchPlatform):

    """Overlay platform to drive system11 machines using a WPC controller."""

    __slots__ = ["delay", "platform", "system11_config", "a_side_queue", "c_side_queue", "debounce_secs",
                 "a_side_done_time", "c_side_done_time", "drivers_holding_a_side", "drivers_holding_c_side",
                 "_a_side_enabled", "_c_side_enabled", "ac_relay_in_transition", "prefer_a_side", "drivers",
                 "relay_switch"]

    def __init__(self, machine: MachineController) -> None:
        """Initialize the board."""
        super().__init__(machine)

        self.delay = DelayManager(machine)

        self.platform = None            # type: Optional[SwitchDriverPlatform]

        self.system11_config = None     # type: Any

        self.a_side_queue = \
            set()   # type: Set[Tuple[DriverPlatformInterface, Optional[PulseSettings], Optional[HoldSettings]]]
        self.c_side_queue = \
            set()   # type: Set[Tuple[DriverPlatformInterface, Optional[PulseSettings], Optional[HoldSettings]]]

        self.a_side_done_time = 0
        self.c_side_done_time = 0
        self.debounce_secs = 0
        self.drivers_holding_a_side = set()     # type: Set[DriverPlatformInterface]
        self.drivers_holding_c_side = set()     # type: Set[DriverPlatformInterface]
        # Internal tracker for which side is enabled, for platforms that don't have switches
        self._a_side_enabled = True
        self._c_side_enabled = False
        self.drivers = {}               # type: Dict[str, DriverPlatformInterface]

        self.ac_relay_in_transition = False
        self.relay_switch = None
        # Specify whether the AC relay should favour the A or C side when at rest.
        # Typically during a game the 'C' side should be preferred, since that is
        # normally where the flashers are which need a quick response without having to wait on the relay.
        # At game over though, it should prefer the 'A' side so that the relay isn't permanently energised.
        self.prefer_a_side = True

    def stop(self):
        """Stop the overlay. Nothing to do here because stop is also called on parent platform."""

    @property
    def a_side_active(self):
        """Return if A side cannot be switched off right away."""
        return self.drivers_holding_a_side or self.a_side_done_time + self.debounce_secs > self.machine.clock.get_time()

    @property
    def a_side_busy(self):
        """Return if A side cannot be switched off right away."""
        return self.a_side_active or self.a_side_queue

    @property
    def a_side_enabled(self):
        """Return true if the A side is currently enabled."""
        if self.ac_relay_in_transition:
            return False
        if self.relay_switch:
            return not self.relay_switch.state
        return self._a_side_enabled

    @property
    def c_side_active(self):
        """Return if C side cannot be switched off right away."""
        return self.drivers_holding_c_side or self.c_side_done_time + self.debounce_secs > self.machine.clock.get_time()

    @property
    def c_side_busy(self):
        """Return if C side cannot be switched off right away."""
        return self.c_side_active or self.c_side_queue

    @property
    def c_side_enabled(self):
        """Return true if the C side is currently enabled."""
        if self.relay_switch:
            # Never enabled if the relay is in transition
            return not self.ac_relay_in_transition and self.relay_switch.state
        return self._c_side_enabled

    async def initialize(self):
        """Automatically called by the Platform class after all the core modules are loaded."""
        # Some platforms (like Fast Retro) may be system11, or may not be.
        # If no system11 config is present, do not initialize the System11 platform
        system11_config = self.machine.config.get('system11')
        if not system11_config:
            return

        # load coil platform
        self.platform = self.machine.get_platform_sections(
            "platform", getattr(system11_config, 'platform', None))

        # we have to wait for coils to be initialized
        self.machine.events.add_handler("init_phase_1", self._initialize)

    def _initialize(self, **kwargs):
        del kwargs
        self._validate_config()

        self.configure_logging('Platform.System11', self.system11_config['console_log'],
                               self.system11_config['file_log'])

        self.log.debug("Configuring A/C Select Relay for driver %s",
                       self.system11_config['ac_relay_driver'])

        self.system11_config['ac_relay_driver'].get_and_verify_hold_power(1.0)
        self.relay_switch = self.system11_config['ac_relay_switch']

        if self.relay_switch:
            self.log.debug("Configuring A/C Select Relay switch %s", self.relay_switch)
            self.relay_switch.add_handler(state=0, callback=self._on_a_side_enabled)
            self.relay_switch.add_handler(state=1, callback=self._on_c_side_enabled)

            # If the platform does not have a physical switch for the AC Relay, a virtual
            # switch may be implemented with a special driver configuration if that
            # driver class has a set_relay() method defined.
            if hasattr(self.system11_config['ac_relay_driver'].hw_driver, 'set_relay'):
                self.system11_config['ac_relay_driver'].hw_driver.set_relay(
                    self.relay_switch.hw_switch,
                    20,  # Ms to delay before reporting closed
                    20   # Ms to delay before reporting open
                )
        self.debounce_secs = self.system11_config['ac_relay_debounce_ms'] / 1000.0
        self.log.debug("Configuring A/C Select Relay transition delay for %sms and debounce for %s",
                       self.system11_config['ac_relay_delay_ms'],
                       self.system11_config['ac_relay_debounce_ms'])

        self.machine.events.add_handler(self.system11_config['prefer_a_side_event'], self._prefer_a_side)
        self.log.info("Configuring System11 driver to prefer A side on event %s",
                      self.system11_config['prefer_a_side_event'])

        self.machine.events.add_handler(self.system11_config['prefer_c_side_event'], self._prefer_c_side)
        self.log.info("Configuring System11 driver to prefer C side on event %s",
                      self.system11_config['prefer_c_side_event'])

    def _prefer_a_side(self, **kwargs):
        del kwargs
        self.prefer_a_side = True
        self._enable_a_side()

    def _prefer_c_side(self, **kwargs):
        del kwargs
        self.prefer_a_side = False
        self._enable_c_side()

    def _validate_config(self):
        self.system11_config = self.machine.config_validator.validate_config(
            'system11', self.machine.config.get('system11', {}))

    def tick(self):
        """System11 main loop.

        Called based on the timer_tick event.
        """
        if self.prefer_a_side:
            if self.a_side_queue:
                self._service_a_side()
            elif self.c_side_queue:
                self._service_c_side()
            elif self.c_side_enabled and not self.c_side_active:
                self._enable_a_side()
        else:
            if self.c_side_queue:
                self._service_c_side()
            elif self.a_side_queue:
                self._service_a_side()
            elif self.a_side_enabled and not self.a_side_active:
                self._enable_c_side()

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict):
        """Configure switch on system11 overlay."""
        return self.platform.configure_switch(number, config, platform_config)

    def validate_switch_section(self, switch, config: dict) -> dict:
        """Validate switch config for overlayed platform."""
        # Check for a platform to have a custom switch validation method, but avoid recursion
        # since the platform will (most likely) inherit from System11OverlayPlatform
        if self.platform is not None:
            return self.platform.validate_switch_section(switch, config)
        return SwitchPlatform.validate_switch_section(self, switch, config)

    async def get_hw_switch_states(self):
        """Get initial hardware state."""
        return await self.platform.get_hw_switch_states()

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict):
        """Configure a driver on the system11 overlay.

        Args:
        ----
            config: Driver config dict
            number: Number of the driver.
            platform_settings: Platform specific config.
        """
        assert self.platform is not None
        number_str = str(number)

        if number_str and (number_str.lower().endswith('a') or number_str.lower().endswith('c')):

            side = number_str[-1:].upper()
            number_str = number_str[:-1]

            # only configure driver once
            if number_str not in self.drivers:
                self.drivers[number_str] = self.platform.configure_driver(config, number_str, platform_settings)

            system11_driver = System11Driver(number, self.drivers[number_str], self, side)

            return system11_driver

        return self.platform.configure_driver(config, number_str, platform_settings)

    @staticmethod
    def _check_if_driver_is_capable_for_rule(driver: DriverPlatformInterface):
        """Check if driver is capable for rule and bail out with an exception if not."""
        if isinstance(driver, System11Driver):
            raise AssertionError("Received a request to set a hardware rule for a System11 driver {}. "
                                 "This is not supported.".format(driver))

    def set_pulse_on_hit_and_release_rule(self, enable_switch, coil):
        """Configure a rule for a driver on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_and_release_rule(enable_switch, coil)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch, coil):
        """Configure a rule for a driver on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_and_enable_and_release_rule(enable_switch, coil)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings, eos_switch: SwitchSettings,
                                                      coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Configure a rule for a driver on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_and_release_and_disable_rule(enable_switch, eos_switch, coil, repulse_settings)

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings,
                                                                 coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Configure a rule for a driver on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_and_enable_and_release_and_disable_rule(enable_switch, eos_switch, coil,
                                                                               repulse_settings)

    def set_pulse_on_hit_rule(self, enable_switch, coil):
        """Configure a rule on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_rule(enable_switch, coil)

    def clear_hw_rule(self, switch, coil):
        """Clear a rule for a driver on the system11 overlay."""
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.clear_hw_rule(switch, coil)

    # pylint: disable-msg=too-many-arguments
    def driver_action(self, driver, pulse_settings: Optional[PulseSettings], hold_settings: Optional[HoldSettings],
                      side: str, timed_enable: bool = False):
        """Add a driver action for a switched driver to the queue (for either the A-side or C-side queue).

        Args:
        ----
            driver: A reference to the original platform class Driver instance.
            pulse_settings: Settings for the pulse or None
            hold_settings:Settings for hold or None
            side: Whatever the driver is on A or C side.

        This action will be serviced immediately if it can, or ASAP otherwise.
        """
        if self.prefer_a_side:
            if side == "A":
                self.a_side_queue.add((driver, pulse_settings, hold_settings, timed_enable))
                if not self.ac_relay_in_transition:
                    self._service_a_side()
            elif side == "C":
                self.c_side_queue.add((driver, pulse_settings, hold_settings, timed_enable))
                if not self.ac_relay_in_transition and not self.a_side_busy:
                    self._service_c_side()
            else:
                raise AssertionError("Invalid side {}".format(side))
        else:
            if side == "C":
                # Sometimes it doesn't make sense to queue the C side (flashers) and play them after
                # switching to the A side (coils) and back. If we are on A side or have a queue on
                # the A side, ignore this C side request.
                if (self.a_side_queue or not self.c_side_enabled) and \
                        not self.system11_config['queue_c_side_while_preferred']:
                    return
                self.c_side_queue.add((driver, pulse_settings, hold_settings, timed_enable))
                if not self.ac_relay_in_transition:
                    self._service_c_side()
            elif side == "A":
                self.a_side_queue.add((driver, pulse_settings, hold_settings, timed_enable))
                # Clear the C-side queue to prioritize A-side and get it switched over faster
                if not self.system11_config['queue_c_side_while_preferred']:
                    self.c_side_queue.clear()
                if not self.ac_relay_in_transition and not self.c_side_busy:
                    self._service_a_side()
            else:
                raise AssertionError("Invalid side {}".format(side))

    def _enable_ac_relay(self):
        self.system11_config['ac_relay_driver'].enable()
        self.ac_relay_in_transition = True
        self._a_side_enabled = False
        self._c_side_enabled = False
        # Without a relay switch, use a delay to wait for the relay to enable
        if not self.relay_switch:
            self.delay.add(ms=self.system11_config['ac_relay_delay_ms'],
                           callback=self._on_c_side_enabled,
                           name='enable_ac_relay')

    def _disable_ac_relay(self):
        self.system11_config['ac_relay_driver'].disable()
        self.ac_relay_in_transition = True
        self._a_side_enabled = False
        self._c_side_enabled = False
        # Clear out the C side queue if we don't want to hold onto it for later
        if not self.system11_config['queue_c_side_while_preferred']:
            self.c_side_queue.clear()
        if not self.relay_switch:
            self.delay.add(ms=self.system11_config['ac_relay_delay_ms'],
                           callback=self._on_a_side_enabled,
                           name='disable_ac_relay')

    # -------------------------------- A SIDE ---------------------------------

    def _enable_a_side(self):
        if self.prefer_a_side:
            if not self.a_side_enabled and not self.ac_relay_in_transition:

                if self.c_side_active:
                    self._disable_all_c_side_drivers()
                    self._disable_ac_relay()
                    return

                if self.c_side_enabled:
                    self._disable_ac_relay()

                else:
                    self._on_a_side_enabled()
        else:
            if (not self.ac_relay_in_transition and
                    not self.a_side_enabled and
                    not self.c_side_busy):
                self._disable_ac_relay()

            elif self.a_side_enabled and self.a_side_queue:
                self._service_a_side()

    def _on_a_side_enabled(self):
        self.ac_relay_in_transition = False
        # If A side has no queue and is not preferred, return to C side
        if not self.a_side_queue and not self.prefer_a_side:
            self._enable_c_side()
            return

        self._c_side_enabled = False
        self._a_side_enabled = True
        self._service_a_side()

    def _service_a_side(self):
        if not self.a_side_queue:
            return

        if not self.a_side_enabled:
            self._enable_a_side()
            return

        while self.a_side_queue:
            driver, pulse_settings, hold_settings, timed_enable = self.a_side_queue.pop()

            if timed_enable:
                wait_ms = driver.timed_enable(pulse_settings, hold_settings) or 0
                wait_secs = (pulse_settings.duration + hold_settings.duration + wait_ms) / 1000.0
                self.a_side_done_time = max(self.a_side_done_time, self.machine.clock.get_time() + wait_secs)

            elif hold_settings is None and pulse_settings:
                wait_ms = driver.pulse(pulse_settings) or 0
                wait_secs = (pulse_settings.duration + wait_ms) / 1000.0
                self.a_side_done_time = max(self.a_side_done_time, self.machine.clock.get_time() + wait_secs)

            elif hold_settings and pulse_settings:
                driver.enable(pulse_settings, hold_settings)
                self.drivers_holding_a_side.add(driver)

            else:  # ms == 0
                driver.disable()
                try:
                    self.drivers_holding_a_side.remove(driver)
                except KeyError:
                    pass

    # -------------------------------- C SIDE ---------------------------------

    def _enable_c_side(self):
        if self.prefer_a_side:
            if not self.c_side_enabled and not self.ac_relay_in_transition:

                if self.a_side_active:
                    self._disable_all_a_side_drivers()
                    self._enable_ac_relay()
                    return

                if self.a_side_enabled:
                    self._enable_ac_relay()

                else:
                    self._on_c_side_enabled()
        else:
            if (not self.ac_relay_in_transition and
                    not self.c_side_enabled and
                    not self.a_side_busy):
                self._enable_ac_relay()

            elif self.c_side_enabled and self.c_side_queue:
                self._service_c_side()

    def _on_c_side_enabled(self):
        self.ac_relay_in_transition = False
        # If C side is not preferred and has no queue, return to A side
        if not self.c_side_queue and self.prefer_a_side:
            self._enable_a_side()
            return

        self._a_side_enabled = False
        self._c_side_enabled = True
        self._service_c_side()

    def _service_c_side(self):
        if not self.c_side_queue:
            return

        if self.ac_relay_in_transition or self.a_side_busy:
            return

        if not self.c_side_enabled:
            self._enable_c_side()
            return

        while self.c_side_queue:
            driver, pulse_settings, hold_settings, timed_enable = self.c_side_queue.pop()

            if timed_enable:
                driver.timed_enable(pulse_settings, hold_settings)
                self.c_side_done_time = max(self.c_side_done_time,
                                            self.machine.clock.get_time() + (pulse_settings.duration / 1000.0))

            elif hold_settings is None and pulse_settings:
                driver.pulse(pulse_settings)
                self.c_side_done_time = max(self.c_side_done_time,
                                            self.machine.clock.get_time() + (pulse_settings.duration / 1000.0))
            elif hold_settings and pulse_settings:
                driver.enable(pulse_settings, hold_settings)
                self.drivers_holding_c_side.add(driver)

            else:
                driver.disable()
                try:
                    self.drivers_holding_c_side.remove(driver)
                except KeyError:
                    pass

    def _disable_all_c_side_drivers(self):
        if self.c_side_active:
            for driver in self.drivers_holding_c_side:
                driver.disable()
            self.drivers_holding_c_side = set()
            self.c_side_done_time = 0

    def _disable_all_a_side_drivers(self):
        if self.a_side_active:
            for driver in self.drivers_holding_a_side:
                driver.disable()
            self.drivers_holding_a_side = set()
            self.a_side_done_time = 0

    def validate_coil_section(self, driver, config):
        """Validate coil config for platform."""
        # Check for a platform to have a custom coil validation method, but avoid recursion
        # since the platform will (most likely) inherit from System11OverlayPlatform
        if self.platform is not None:
            return self.platform.validate_coil_section(driver, config)
        return DriverPlatform.validate_coil_section(self, driver, config)


class System11Driver(DriverPlatformInterface):

    """Represent one driver on the system11 overlay.

    Two of those drivers may be created for one real driver. One for the A and one for the C side.
    """

    __slots__ = ["platform_driver", "overlay", "side"]

    def __init__(self, number, platform_driver: DriverPlatformInterface, overlay, side) -> None:
        """Initialize driver."""
        super().__init__(platform_driver.config, number)
        self.platform_driver = platform_driver
        self.overlay = overlay
        self.side = side

    def __repr__(self):
        """Pretty print."""
        return "System11Driver.{}".format(self.number)

    def get_board_name(self):
        """Return name of driver board."""
        return self.platform_driver.get_board_name()

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse driver."""
        self.overlay.driver_action(self.platform_driver, pulse_settings, None, self.side)

        # Usually pulse() returns the value (in ms) that the driver will pulse
        # for so we can update Driver.time_when_done. But with A/C switched
        # coils, we don't know when exactly that will be, so we return -1
        return -1

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable driver."""
        self.overlay.driver_action(self.platform_driver, pulse_settings, hold_settings, self.side)

    def disable(self):
        """Disable driver."""
        self.overlay.driver_action(self.platform_driver, None, None, self.side)

    def timed_enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Pulse and enable the coil for an explicit duration."""
        self.overlay.driver_action(self.platform_driver, pulse_settings, hold_settings, self.side, True)



"""Text stack entry support class for segment displays."""
from typing import Optional, List

from mpf.core.rgb_color import RGBColor
from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType


# pylint: disable=too-many-instance-attributes,too-many-arguments,too-few-public-methods
class TextStackEntry:

    """An entry in the text stack for a segment display."""

    __slots__ = ["text", "colors", "flashing", "flash_mask", "transition", "transition_out", "priority", "key"]

    def __init__(self, text: str, color: Optional[List[RGBColor]],
                 flashing: Optional[FlashingType] = None, flash_mask: Optional[str] = None,
                 transition: Optional[dict] = None, transition_out: Optional[dict] = None,
                 priority: int = 0, key: str = None):
        """Class initializer."""
        self.text = text
        self.colors = color
        self.flashing = flashing
        self.flash_mask = flash_mask
        self.transition = transition
        self.transition_out = transition_out
        self.priority = priority
        self.key = key

    def __repr__(self):
        """Return str representation."""
        return '<TextStackEntry: {} (priority: {}, key: {} colors: {}) >'.format(
            self.text, self.priority, self.key, self.colors)



"""MPF plugin which adds events from monitoring a Twitch chat room."""
import threading
from mpf.core.plugin import MpfPlugin

from .twitch.twitch_client import TwitchClient

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class TwitchBot(MpfPlugin):

    """Adds Twitch chat room events."""

    __slots__ = ["client"]
    config_section = 'twitch_client'

    def __init__(self, *args, **kwargs):
        """Initialize class variables."""
        super().__init__(*args, **kwargs)
        self.client = None

    def initialize(self):
        """Initialize Twitch client."""
        self.configure_logging(self.name)
        self.config = self.machine.config_validator.validate_config(
            "twitch_client", self.machine.config['twitch_client'])

        self.log.info('Attempting to connect to Twitch')

        user_var = self.config['user_var']
        password_var = self.config['password_var']
        channel_var = self.config['channel_var']

        user = self.machine.variables.get_machine_var(
            user_var
        ) if user_var is not None else self.config['user']

        password = self.machine.variables.get_machine_var(
            password_var
        ) if password_var is not None else self.config['password']

        channel = self.machine.variables.get_machine_var(
            channel_var
        ) if channel_var is not None else self.config['channel']

        self.client = TwitchClient(self.machine, user, password, channel,
                                   self.machine.clock.loop)
        thread = threading.Thread(target=self.client.start, args=())
        thread.daemon = True
        thread.start()

        if self.client.is_connected():
            self.info_log('Successful connection to Twitch')
        else:
            self.info_log('Connecting...')



"""A speedometer to measure the speed of a ball."""
from mpf.core.system_wide_device import SystemWideDevice


class Speedometer(SystemWideDevice):

    """A device which represents a tachometer."""

    config_section = 'speedometers'
    collection = 'speedometers'
    class_label = 'speedometer'

    __slots__ = ["time_start"]

    def __init__(self, machine, name):
        """Initialize speedometer."""
        super().__init__(machine, name)
        self.time_start = None

    async def device_added_system_wide(self):
        """Register switch handlers on load."""
        await super().device_added_system_wide()

        self.machine.switch_controller.add_switch_handler_obj(
            self.config['start_switch'], self._handle_start_switch, 1)

        self.machine.switch_controller.add_switch_handler_obj(
            self.config['stop_switch'], self._handle_stop_switch, 1)

    def _handle_start_switch(self, **kwargs):
        del kwargs
        self.time_start = self.config['start_switch'].last_change

    def _handle_stop_switch(self, **kwargs):
        del kwargs
        if self.time_start is not None:
            delta = self.config['stop_switch'].last_change - self.time_start
            self.time_start = None
            print(delta)
            self.machine.events.post("{}_hit".format(self.name), delta=delta)
            # TODO: post event



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# sphinx-doc config file

from copy import copy
import os
import re
import shutil
import sys
import time

import git
import sphinx_rtd_theme
import stat

from mpf.core.config_validator import ConfigValidator

sys.path.insert(0, os.path.abspath(os.pardir))

from mpf.core.utility_functions import Util
import mpf._version
from mpf.core.config_processor import ConfigProcessor


# MPF Examples Link -------------------------------------------------------

mpf_examples_link_name = 'examples'


def setup_mpf_examples_link():

    if os.path.isdir(os.path.join(os.getcwd(), os.pardir, os.pardir, 'mpf-examples')):
        examples_root = os.path.join(os.getcwd(), os.pardir, os.pardir, 'mpf-examples')

    elif os.path.isdir(os.path.join(sys.prefix, 'src', 'mpf-examples')):
        examples_root = os.path.join(sys.prefix, 'src', 'mpf-examples')

    else:
        raise RuntimeError("Cannot find mpf-examples repo. Aborting")

    verify_version(os.path.join(examples_root, '_version.py'))

    # See if symlink already exists, and goes to the right place.
    exists = True
    bad = True
    try:
        linkinfo = os.stat(mpf_examples_link_name, follow_symlinks=False)
        if (stat.S_ISLNK(linkinfo.st_mode)):
            link_path = os.readlink(mpf_examples_link_name)
            # This sometimes has trouble on Windows because realpath doesn't canonise case of the drive letter
            if os.path.realpath(link_path) == os.path.realpath(examples_root):
                bad = False
    except FileNotFoundError:
        exists = False

    # If it exists, but isn't a link or is to the wrong place, delete it
    if bad and exists:
        os.unlink(mpf_examples_link_name)

    # If it didn't exist or we just deleted it, created it
    if bad or not exists:
        # Flush to ensure message is displayed before potential system error from symlink failing
        print("Creating '{}' link to {}".format(mpf_examples_link_name, examples_root), flush=True)
        os.symlink(examples_root, mpf_examples_link_name)

def verify_version(version_file):

    #  http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
    verstrline = open(version_file, "rt").read()
    VSRE = r"^mpf_version_required = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        mpf_version_required_string = mo.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s." % (version_file,))

    if mpf_version_required_string != mpf._version.__short_version__:
        raise RuntimeError("mpf-examples version mismatch. MPF is version {} "
                           "but the mpf-examples repo found requires MPF {}".format(
            mpf._version.__short_version__, mpf_version_required_string))

setup_mpf_examples_link()


# -- General configuration ------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinxcontrib.napoleon',
    'sphinx.ext.autosummary',
    'sphinxcontrib.youtube',
]

templates_path = ['_templates']

source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# The short X.Y version.
version = mpf._version.__short_version__
# The full version, including alpha/beta/rc tags.
release = mpf._version.__version__

# General information about the project.
project = 'Mission Pinball Framework v{} Developer Documentation'.format(version)
copyright = '2013-%s, The Mission Pinball Framework Team' % time.strftime('%Y')
author = 'The Mission Pinball Framework Team'

# dev warning box will be shown in HTML builds for the following GitHub branch
# names:
branches_for_dev_warning = ['dev']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

todo_include_todos = True

numpydoc_show_inherited_class_members = False

autodoc_default_flags = [
         # Make sure that any autodoc declarations show the right members
         "members",
         "show-inheritance",
]

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.
# "<project> v<release> documentation" by default.
# html_title = 'Mission Pinball Framework API v{}'.format(version)

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = 'MPF API v{}'.format(version)

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or
# 32x32
# pixels large.
html_favicon = '_static/images/icons/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not None, a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# The empty string is equivalent to '%b %d, %Y'.
html_last_updated_fmt = '%b %d, %Y'

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr', 'zh'
# html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# 'ja' uses this config value.
# 'zh' user can custom change `jieba` dictionary path.
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
# html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'MPFAPIdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',

    # Latex figure (float) alignment
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'MPFdev.tex', 'MPF Documentation Developer Documentation',
     'The Mission Pinball Framework Team', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'mpf-dev', 'MPF Developer Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False

# -- Show warnings for dev branches in HTML docs --------------------------

# If this is running on ReadTheDocs.org, the context dict will be overwritten
# by theirs which will contain the propoer branch name (since the git command
# doesn't work there.

context = dict()

try:
    context['github_version'] = git.Repo('..').active_branch.name
except Exception:       # noqa
    context['github_version'] = None


def setup(app):
    app.add_stylesheet('mpf.css')
    RstBuilder().build_rst_files()

    # We need to do this in the setup() function since ReadTheDocs will append
    # the context dict to the end of conf.py which means we don't have the
    # populated value at the global context yet, so we need to do it here.

    if globals()['context']['github_version'] in branches_for_dev_warning:

        globals()['rst_prolog'] = '''

        .. only:: html

           .. warning::

              **This documentation is for an unreleased version of MPF!**

              This is the developer documentation for MPF |version|, which is
              the "dev" (next) release of MPF that is a work-in-progress. Use
              the "Read the Docs" link in the lower left corner to view the
              developer docs for the version of MPF you're using.

        '''


class RstBuilder(object):

    """Builds Rst Config."""

    def __init__(self):

        self.dest_folder = 'api'
        self.template_folder = '_templates'

        config_processor = ConfigProcessor(False, False)
        self.mpfconfig = config_processor.load_config_files_with_cache(
            [os.path.join(os.pardir, 'mpf', 'mpfconfig.yaml')], 'machine', config_spec=config_processor.load_config_spec())

        self.doc_sections = dict()
        self.index_entries = dict()
        self.templates = dict()

        self.populate_doc_sections()
        self.get_templates()

        self.additional_files = list()  # section, title, file
        self.additional_files.append(('machine', 'machine', 'self.machine'))

        # add manually-created files to the index entries so they get included
        # in the index
        # self.doc_sections['machine'].append(('machine', 'self.machine'))

    def populate_doc_sections(self):

        self.doc_sections['machine'] = self.mpfconfig['mpf']['core_modules']

        self.doc_sections['config_players'] = dict()

        for name, module_ in self.mpfconfig['mpf']['config_players'].items():
            self.doc_sections['config_players'
                              ]['{}_player'.format(name)] = module_

        self.doc_sections['platforms'] = self.mpfconfig['mpf']['platforms']

        for plugin in Util.string_to_event_list(self.mpfconfig['mpf']['plugins']):
            name = plugin.split('.')[-2]
            self.doc_sections['machine'][name] = plugin

        self.doc_sections['devices'] = dict()

        for device in self.mpfconfig['mpf']['device_modules']:
            device_cls = Util.string_to_class(device)
            name = device_cls.collection
            self.doc_sections['devices'][name] = device

        self.doc_sections['modes'] = dict()

        for name in [x for x in
                     os.walk(os.path.join(os.pardir, 'mpf', 'modes'))][0][1]:

            if name.startswith('__'):
                continue

            class_name = ''.join([x.capitalize() for x in name.split('_')])

            self.doc_sections['modes'][name] = (
                'mpf.modes.{0}.code.{0}.{1}'.format(name, class_name))

        self.doc_sections['tests'] = dict(
            MockBcpClient='mpf.tests.MpfBcpTestCase.MockBcpClient',
            MpfBcpTestCase='mpf.tests.MpfBcpTestCase.MpfBcpTestCase',
            MpfFakeGameTestCase='mpf.tests.MpfFakeGameTestCase.MpfFakeGameTestCase',
            MpfGameTestCase='mpf.tests.MpfGameTestCase.MpfGameTestCase',
            MpfMachineTestCase='mpf.tests.MpfMachineTestCase.MpfMachineTestCase',
            MpfTestCase='mpf.tests.MpfTestCase.MpfTestCase',
            TestMachineController='mpf.tests.MpfTestCase.TestMachineController',
            TestDataManager='mpf.tests.TestDataManager.TestDataManager',
        )

        self.doc_sections['misc'] = dict(
            data_manager='mpf.core.data_manager.DataManager',
            delay_manager='mpf.core.delays.DelayManager',
            delay_manager_registry='mpf.core.delays.DelayManagerRegistry',
            LogMixin='mpf.core.logging.LogMixin',
            Players='mpf.core.player.Player',
            RGBColor='mpf.core.rgb_color.RGBColor',
            RGBAColor='mpf.core.rgba_color.RGBAColor',
            Randomizer='mpf.core.randomizer.Randomizer',
            Timers='mpf.core.timer.Timer',
        )

        self.doc_sections['misc']['Ball Search'] = 'mpf.core.ball_search.BallSearch'
        self.doc_sections['misc']['File Manager'] = 'mpf.core.file_manager.FileManager'
        self.doc_sections['misc']['Utility Functions'] = 'mpf.core.utility_functions.Util'
        self.doc_sections['misc']['Mode base class'] = 'mpf.core.mode.Mode'

        # populate the index entries

        for name in self.doc_sections.keys():
            self.index_entries[name] = list()

    def get_templates(self):

        for suffix in ('', '_overview'):

            for name in self.doc_sections.keys():
                with open(os.path.join(
                        templates_path[0], '{}{}.rst'.format(name, suffix)), 'r', encoding="utf-8") as f:
                    self.templates[name + suffix] = f.read()

    def build_rst_files(self):
        print('Building RST files from templates')

        try:
            shutil.rmtree(self.dest_folder)
        except FileNotFoundError:
            pass

        os.makedirs(self.dest_folder)

        for section, items in self.doc_sections.items():
            for name, module_ in items.items():
                self.create_rst_file(section, name, module_)

        self.handle_additional_files()
        self.write_overviews()

        shutil.copyfile(os.path.join(self.template_folder, 'index.rst'),
                 os.path.join(self.dest_folder, 'index.rst'))

    def handle_additional_files(self):
        for section, title, file in self.additional_files:
            self.index_entries[section].append((title, file))

            file_name = '{}.rst'.format(file)

            shutil.copyfile(os.path.join(self.template_folder, file_name),
                 os.path.join(self.dest_folder, file_name))

    def create_rst_file(self, section, name, module_):
        this_rst = copy(self.templates[section])

        file_name = this_rst.split('\n')[0].format(name=name)

        file_name = file_name.replace('[', '.')

        for char in "''*] ":
            file_name = file_name.replace(char, '')

        if file_name[-1] == '.':
            file_name = file_name[:-1]

        with open(os.path.join(self.dest_folder, '{}.rst'.format(file_name)),
                  'w') as f:
            f.write(this_rst.format(
                name=name,
                module_underline='=' * (len(this_rst.split('\n')[0]) +
                                        len(name) - 6),
                full_path_to_class=module_,
                article='an' if name[0] in ('a', 'e', 'i', 'o', 'u') else 'a',
                cap_name=name.capitalize()))

        self.index_entries[section].append((name, file_name))

    def write_overviews(self):
        string_indices = dict()
        for section, files in self.index_entries.items():
            string_indices[section] = self.create_rst_file_list(files)

            template = self.templates['{}_overview'.format(section)].format(
                **string_indices)

            file_name = os.path.join(self.dest_folder,
                                     '{}_overview.rst'.format(section))

            with open(file_name, 'w') as f:
                f.write(template)

    @staticmethod
    def create_rst_file_list(file_list):
        file_list.sort()

        final_string = str()

        for name, file in file_list:
            final_string += '   {0} <{1}>\n'.format(name, file)

        return final_string



"""PKONE Lights."""
import logging
from collections import namedtuple

from mpf.core.platform import LightConfig
from mpf.core.platform_batch_light_system import PlatformBatchLight, PlatformBatchLightSystem
from mpf.platforms.interfaces.light_platform_interface import LightPlatformSoftwareFade

MYPY = False
if MYPY:  # pragma: no cover
    from mpf.platforms.pkone.pkone import PKONEHardwarePlatform  # pylint: disable-msg=cyclic-import,unused-import

PKONESimpleLEDNumber = namedtuple("PKONESimpleLEDNumber", ["board_address_id", "led_number"])


class PKONESimpleLED(LightPlatformSoftwareFade):

    """A simple led (single emitter/color) on a PKONE Extension board. Simple leds are either on or off."""

    __slots__ = ["log", "send", "platform"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, number: PKONESimpleLEDNumber, sender, platform: "PKONEHardwarePlatform") -> None:
        """Initialize light."""
        super().__init__(number, platform.machine.clock.loop, 0)
        self.log = logging.getLogger('PKONESimpleLED')
        self.send = sender
        self.platform = platform

    def set_brightness(self, brightness: float) -> None:
        """Turn on or off this Simple LED.

        Args:
        ----
            brightness: brightness 0 (off) to 255 (on) for this Simple LED. PKONE Simple LEDs only
            support on (>0) or off.
        """
        # send the new simple LED state command (on or off)
        cmd = "PLS{}{:02d}{}".format(self.number.board_address_id,
                                     self.number.led_number,
                                     1 if brightness > 0 else 0)
        self.log.debug("Sending Simple LED Control command: %s", cmd)
        self.send(cmd)

    def get_board_name(self):
        """Return PKONE Lightshow addr."""
        if self.number.board_address_id not in self.platform.pkone_extensions.keys():
            return "PKONE Unknown Board"
        return "PKONE Lightshow Board {}".format(self.number.board_address_id)

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        raise AssertionError("Not possible using Simple LEDs on PKONE Lightshow board.")

    def get_successor_number(self):
        """Return next number."""
        raise AssertionError("Not possible using Simple LEDs on PKONE Lightshow board.")

    def __lt__(self, other):
        """Order lights by board and led number."""
        return (self.number.board_address_id, self.number.led_number) < (
            other.number.board_address_id, other.number.led_number)


class PKONELEDChannel(PlatformBatchLight):

    """Represents a single LED channel connected to a PKONE hardware platform Lightshow board."""

    __slots__ = ["board_address_id", "group", "index", "config", "_hardware_aligned"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, board_address_id, group, index,
                 config: LightConfig, light_system: PlatformBatchLightSystem) -> None:
        """Initialize LED."""
        super().__init__("{}-{}-{}".format(board_address_id, group, index), light_system)
        self.board_address_id = int(board_address_id)
        self.group = int(group)
        self.index = int(index)
        self.config = config
        self._hardware_aligned = False

    def set_hardware_aligned(self, hardware_aligned: bool = True):
        """Set whether or not this channel is aligned to hardware boundaries."""
        self._hardware_aligned = hardware_aligned

    def get_max_fade_ms(self) -> int:
        """Return max fade time."""
        return 40960 if self._hardware_aligned else 0

    def get_board_name(self):
        """Return the board of this light."""
        return "PKONE LED Channel {} on Lightshow Board (Address ID {}, Group {}, Light: {}, " \
               "Hardware Aligned: {})".format(self.index,
                                              self.board_address_id,
                                              self.group,
                                              self.config.name,
                                              "Yes" if self._hardware_aligned else "No")

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        return self.board_address_id == other.board_address_id and self.group == other.group and \
            self.index == other.index + 1

    def get_successor_number(self):
        """Return next number."""
        return "{}-{}-{}".format(self.board_address_id, self.group, self.index + 1)

    def get_predecessor_number(self):
        """Return previous number."""
        if self.index > 0:
            return "{}-{}-{}".format(self.board_address_id, self.group, self.index - 1)
        return None

    def __lt__(self, other):
        """Order lights by their order on the hardware."""
        return (self.board_address_id, self.group, self.index) < (other.board_address_id, other.group, other.index)

    def __repr__(self):
        """Return str representation."""
        return "<PKONELEDChannel board_address_id={} group={} index={}>".format(
            self.board_address_id, self.group, self.index)



"""Switch test for P3."""
import sys

try:    # pragma: no cover
    import pinproc
    pinproc_imported = True
except ImportError:     # pragma: no cover
    try:
        if sys.platform == 'darwin':
            from mpf.platforms.pinproc.osx import pinproc
        elif sys.platform == 'win32':
            if sys.platform.architecture()[0] == '32bit':
                from mpf.platforms.pinproc.x86 import pinproc
            elif sys.platform.architecture()[0] == '64bit':
                from mpf.platforms.pinproc.x64 import pinproc

        pinproc_imported = True

    except ImportError:
        pinproc_imported = False
        pinproc = None

proc = pinproc.PinPROC(pinproc.normalize_machine_type('pdb'))

for switch, state in enumerate(proc.switch_get_states()):

    if not switch % 16:
        print('')
        print('SW-16 Board Address {0}'.format(switch / 16))

    if state != 3:
        print('Switch {0} State {1}'.format(switch, state))



"""PKONE Extension board."""
import logging


# pylint: disable-msg=too-few-public-methods
class PKONEExtensionBoard:

    """PKONE Extension board."""

    __slots__ = ["log", "addr", "firmware_version", "hardware_rev", "switch_count",
                 "coil_count", "servo_count"]

    def __init__(self, addr, firmware_version, hardware_rev):
        """Initialize PKONE Extension board."""
        self.log = logging.getLogger('PKONEExtensionBoard {}'.format(addr))
        self.addr = addr
        self.firmware_version = firmware_version
        self.hardware_rev = hardware_rev
        self.switch_count = 35  # numbers 1 - 35 (31-35 are NC opto switches)
        self.coil_count = 10  # numbers 1 - 10
        self.servo_count = 4    # numbers 11-14

    def get_description_string(self) -> str:
        """Return description string."""
        return "PKONE Extension Board {} - Firmware: {}, Hardware Rev: {}, " \
               "Switches: {}, Coils: {}, Servos: {}".format(
                   self.addr,
                   self.firmware_version,
                   self.hardware_rev,
                   self.switch_count,
                   self.coil_count,
                   self.servo_count)



"""Test score reels."""
import sys
import unittest
from unittest.mock import MagicMock

from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase

'''
These tests were disabled for Python 3.11 and newer by Brian (toomanybrians) on Feb 28, 2023.

It's only the score reel tests that fail, I think based on the asyncio changes in 3.11 and how the
score_reel device starts up. I do not have physical score reels to test with. I assume this is not a
hard problem to solve for someone who knows what they're doing. I spent too much time already. Good luck! :)

The failure error is a red herring. It talks about MpfTestCase.advance_time_and_run() and needing a task
to be passed instead of a coroutine. But that's not the problem as even wrapping the coro in a task it still
fails.

'''

class TestScoreReels(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/score_reels/'

    def _synchronise_to_reel(self):
        pass

    def testOvershooting(self):
        player1_10k = self.machine.coils["player1_10k"].hw_driver
        player1_1k = self.machine.coils["player1_1k"].hw_driver
        player1_100 = self.machine.coils["player1_100"].hw_driver
        player1_10 = self.machine.coils["player1_10"].hw_driver
        chime1 = self.machine.coils["chime1"].hw_driver
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)
        chime1.pulse = MagicMock(return_value=10)

        self._synchronise_to_reel()
        # reel desyncs while idle
        self.release_switch_and_run("score_1p_10_0", 0)

        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(0, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # resynced
        self.hit_switch_and_run("score_1p_10_0", 0)
        self.advance_time_and_run(10)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(0, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # advance one position
        self.machine.score_reels["score_1p_100"].set_destination_value(1)
        self.advance_time_and_run(.02)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # success
        self.release_switch_and_run("score_1p_100_0", 0)
        self.advance_time_and_run(10)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # the reel jumps back (for some reason)
        self.hit_switch_and_run("score_1p_100_0", 0)
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(2, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # resync
        self.release_switch_and_run("score_1p_100_0", 0)
        self.advance_time_and_run(10)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(2, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

    def testScoring(self):
        player1_10k = self.machine.coils["player1_10k"].hw_driver
        player1_1k = self.machine.coils["player1_1k"].hw_driver
        player1_100 = self.machine.coils["player1_100"].hw_driver
        player1_10 = self.machine.coils["player1_10"].hw_driver
        chime1 = self.machine.coils["chime1"].hw_driver
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)
        chime1.pulse = MagicMock(return_value=10)

        self.start_game()
        self.assertGameIsRunning()
        self.assertPlayerNumber(1)
        self.assertBallNumber(1)

        self.advance_time_and_run()
        self._synchronise_to_reel()
        self.machine.game.player.score = 110
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.hit_switch_and_run("score_1p_10k_0", 0)
        self.hit_switch_and_run("score_1p_1k_0", 0)
        self.release_switch_and_run("score_1p_100_0", 0)
        self.release_switch_and_run("score_1p_10_0", 0)

        self.advance_time_and_run(10)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.assertEqual(0, chime1.pulse.call_count)
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)

        self._synchronise_to_reel()
        self.machine.game.player.score = 11207
        self.advance_time_and_run(.005)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(0, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        self.advance_time_and_run(.02)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        self.advance_time_and_run(.02)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(1, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        self.advance_time_and_run(.02)
        self.assertEqual(1, player1_10k.pulse.call_count)
        self.assertEqual(1, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        self.release_switch_and_run("score_1p_10k_0", 0)
        self.release_switch_and_run("score_1p_1k_0", 0)

        self.advance_time_and_run(.17)
        self.assertEqual(2, player1_10.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(1, chime1.pulse.call_count)
        self.assertEqual(3, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(4, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(5, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(6, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(7, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(1, player1_10k.pulse.call_count)
        self.assertEqual(1, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(8, player1_10.pulse.call_count)

        # it was stuck somewhere before 9
        self.advance_time_and_run(.2)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(9, player1_10.pulse.call_count)

        self.advance_time_and_run(.1)
        self.hit_switch_and_run("score_1p_10_9", 0)
        self.advance_time_and_run(.1)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(10, player1_10.pulse.call_count)

        self.advance_time_and_run(.1)
        self.release_switch_and_run("score_1p_10_9", 0)
        self.hit_switch_and_run("score_1p_10_0", 0)
        self.advance_time_and_run(.1)

        # no more changes
        self.advance_time_and_run(10)
        self.assertEqual(1, player1_10k.pulse.call_count)
        self.assertEqual(1, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(10, player1_10.pulse.call_count)
        self.assertEqual(1, chime1.pulse.call_count)

    def testAdvanceingFailure(self):
        player1_10k = self.machine.coils["player1_10k"].hw_driver
        player1_1k = self.machine.coils["player1_1k"].hw_driver
        player1_100 = self.machine.coils["player1_100"].hw_driver
        player1_10 = self.machine.coils["player1_10"].hw_driver
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)
        self.start_game()

        self._synchronise_to_reel()
        self.machine.game.player.score += 110
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.release_switch_and_run("score_1p_10_0", 0)
        # switch for pos 0 stays on

        # it retries
        self.advance_time_and_run(.25)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(2, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # and again
        self.advance_time_and_run(.25)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(3, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        # finally
        self.release_switch_and_run("score_1p_100_0", 0)

        # no more retries
        self.advance_time_and_run(10)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(3, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

    def testThreePlayers(self):
        player1_10k = self.machine.coils["player1_10k"].hw_driver
        player1_1k = self.machine.coils["player1_1k"].hw_driver
        player1_100 = self.machine.coils["player1_100"].hw_driver
        player1_10 = self.machine.coils["player1_10"].hw_driver
        player2_10 = self.machine.coils["player2_10"].hw_driver
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)
        player2_10.pulse = MagicMock(return_value=10)
        self.start_game()

        # add two more players
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertEqual(3, self.machine.game.num_players)

        self._synchronise_to_reel()
        self.machine.game.player.score += 110
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.release_switch_and_run("score_1p_10_0", 0)
        self.release_switch_and_run("score_1p_100_0", 0)

        # no more fires
        self.advance_time_and_run(10)
        self.machine_run()
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # drain ball
        self.drain_all_balls()
        self.advance_time_and_run(.1)
        self.assertEqual(2, self.machine.game.player.number)

        self.machine.game.player.score += 20
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.assertEqual(1, player2_10.pulse.call_count)
        self.release_switch_and_run("score_2p_10_0", 0)

        self.advance_time_and_run(.3)
        self.machine_run()
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        # no more changes
        self.advance_time_and_run(10)
        self.machine_run()
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        # drain ball
        self.drain_all_balls()
        self.advance_time_and_run(.15)
        self.assertEqual(3, self.machine.game.player.number)

        # player3 reuses the reels from player 1. machine resets them
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(7, player1_100.pulse.call_count)
        self.assertEqual(7, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        for i in range(2):
            self.advance_time_and_run(.2)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(9, player1_100.pulse.call_count)
        self.assertEqual(9, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        self.hit_switch_and_run("score_1p_10_9", 0)
        self.hit_switch_and_run("score_1p_100_9", 0)

        self.advance_time_and_run(.2)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(10, player1_100.pulse.call_count)
        self.assertEqual(10, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        self.hit_switch_and_run("score_1p_10_0", 0)
        self.hit_switch_and_run("score_1p_100_0", 0)
        self.release_switch_and_run("score_1p_10_9", 0)
        self.release_switch_and_run("score_1p_100_9", 0)

        self.advance_time_and_run(10)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(10, player1_100.pulse.call_count)
        self.assertEqual(10, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)


class TestScoreReelsVirtual(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_platform(self):
        return "smart_virtual"

    def get_machine_path(self):
        return 'tests/machine_files/score_reels/'

    def testScoringVirtual(self):
        self.machine.playfield.add_ball = MagicMock()
        self.machine.ball_controller.num_balls_known = 3
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")

        # wait for reset of score reels
        self.advance_time_and_run(4)
        self.assertGameIsRunning()
        self.assertPlayerNumber(1)

        self.advance_time_and_run()
        self.assertSwitchState("score_1p_10_0", 1)
        self.assertSwitchState("score_1p_100_0", 1)
        self.assertSwitchState("score_1p_10_9", 0)
        self.assertSwitchState("score_1p_100_9", 0)

        self.machine.game.player.score += 110
        self.advance_time_and_run(10)

        self.assertSwitchState("score_1p_10_0", 0)
        self.assertSwitchState("score_1p_100_0", 0)
        self.assertSwitchState("score_1p_10_9", 0)
        self.assertSwitchState("score_1p_100_9", 0)

        self.machine.game.player.score = 990
        self.advance_time_and_run(10)

        self.assertSwitchState("score_1p_10_0", 0)
        self.assertSwitchState("score_1p_100_0", 0)
        self.assertSwitchState("score_1p_10_9", 1)
        self.assertSwitchState("score_1p_100_9", 1)

        self.machine.game.player.score = 1900
        self.advance_time_and_run(10)

        self.assertSwitchState("score_1p_10_0", 1)
        self.assertSwitchState("score_1p_100_0", 0)
        self.assertSwitchState("score_1p_10_9", 0)
        self.assertSwitchState("score_1p_100_9", 1)

    def test_inactive_player_scoring(self):
        self.start_two_player_game()
        self.assertGameIsRunning()
        self.assertPlayerNumber(1)

        self.advance_time_and_run()
        self.assertEqual([None, 0, 0, 0, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 0], self.machine.score_reel_groups["player2"].desired_value_list)

        # increasing score for player two should only affect score of that player
        self.machine.game.player_list[1].score = 3333
        self.advance_time_and_run()
        self.assertEqual([None, 0, 0, 0, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 3], self.machine.score_reel_groups["player2"].desired_value_list)

        self.machine.game.player_list[0].score = 2222
        self.advance_time_and_run()
        self.assertEqual([None, 2, 2, 2, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 3], self.machine.score_reel_groups["player2"].desired_value_list)

        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.assertBallNumber(1)

        self.machine.game.player_list[0].score = 4444
        self.advance_time_and_run()
        self.assertEqual([None, 4, 4, 4, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 3], self.machine.score_reel_groups["player2"].desired_value_list)

        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.assertBallNumber(2)

        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.assertBallNumber(2)

        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.assertBallNumber(3)

        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.assertBallNumber(3)

        self.drain_all_balls()
        self.assertGameIsNotRunning()

        # start game. we cannot use start_two_player_game because score reels block the start for some seconds
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsRunning()
        self.assertEqual(0, self.machine.game.num_players)
        self.advance_time_and_run(5)
        self.assertEqual(1, self.machine.game.num_players)
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run(1)
        self.assertEqual(2, self.machine.game.num_players)
        self.assertPlayerNumber(1)

        self.assertEqual([None, 0, 0, 0, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 0], self.machine.score_reel_groups["player2"].desired_value_list)



"""MPF plugin which connects segment displays to MPF-MC to update segment display emulator widgets."""

from mpf.core.plugin import MpfPlugin
from mpf.core.rgb_color import RGBColor
from mpf.devices.segment_display.segment_display_text import ColoredSegmentDisplayText
from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class VirtualSegmentDisplayConnector(MpfPlugin):

    """MPF plugin which connects segment displays to MPF-MC to update segment display emulator widgets."""

    __slots__ = ["bcp_client"]

    config_section = 'virtual_segment_display_connector'

    def __init__(self, machine):
        """Initialize virtual segment display connector plugin."""
        super().__init__(machine)
        self.bcp_client = None

        if self.config_section in self.machine.config and not self.machine.bcp.enabled:
            self.machine.log.debug('Disabling virtual_segment_display_connector because BCP is disabled.')

    @property
    def is_plugin_enabled(self):
        """Return true if this plugin is enabled."""
        return self.config_section in self.machine.config and self.machine.bcp.enabled

    def initialize(self):
        """Initialize the virtual segment display connector."""
        self.configure_logging(self.name)
        self.config = self.machine.config_validator.validate_config(
            "virtual_segment_display_connector", self.machine.config['virtual_segment_display_connector'])

        # All BCP clients should be connected (plugins are loaded after BCP is initialized).
        # Determine which connection to use to communicate with MPF-MC
        self.bcp_client = self.machine.bcp.transport.get_named_client(self.config['bcp_connection'])
        if not self.bcp_client:
            raise AssertionError("Could not establish BCP connection to MPF-MC via using client name {}.".format(
                self.config['bcp_connection']))

        # loop over all configured segment displays adding the connector to each one
        if 'segment_displays' in self.config:
            for display in self.config['segment_displays']:
                display.add_virtual_connector(self)

    # pylint: disable=too-many-arguments
    def set_text(self, name: str, text: ColoredSegmentDisplayText, flashing: FlashingType, flash_mask: str) -> None:
        """Set the display text to send to MPF-MC via BCP."""
        colors = text.get_colors()
        self.machine.bcp.interface.bcp_trigger_client(
            client=self.bcp_client,
            name='update_segment_display',
            segment_display_name=name,
            text=text.convert_to_str(),
            flashing=str(flashing.value),
            flash_mask=flash_mask,
            colors=[RGBColor(color).hex for color in colors] if colors else None)



"""Command to initialize an MPF project folder."""

import os
import pathlib
import shutil

SUBCOMMAND = True
INITIAL_MODES = ("attract", "base")

class Command:

    """Initializes an MPF project folder."""

    __slots__ = ("path", )

    def __init__(self, argv, path):
        """Generate config and mode folders."""
        # If a path is provided, use it
        if len(argv) > 1:
            path = os.path.join(path, argv[1])
        self.path = path

        print(f"Initializing MPF project at path {self.path}")

        try:
            os.stat(self.path)
        except FileNotFoundError:
            os.makedirs(self.path)

        self.generate_config_file()
        self.generate_mode_folder()

        print("MPF project initialization complete!")

    def generate_config_file(self):
        # Create the config folder
        config_dir = os.path.join(self.path, "config")
        try:
            os.stat(config_dir)
        except FileNotFoundError:
            os.mkdir(config_dir)

        config_file = os.path.join(config_dir, "config.yaml")
        try:
            os.stat(config_file)
        except FileNotFoundError:
            base_config_file = os.path.join(pathlib.Path(__file__).parent.absolute(), "templates/config.yaml")
            shutil.copy2(base_config_file, config_file)

    def generate_mode_folder(self):
        mode_dir = os.path.join(self.path, "modes")
        try:
            os.stat(mode_dir)
        except FileNotFoundError:
            os.mkdir(mode_dir)


        for mode in INITIAL_MODES:
            mode_path = os.path.join(mode_dir, mode, "config")
            try:
                os.stat(mode_path)
            except FileNotFoundError:
                os.makedirs(mode_path)
            mode_config = os.path.join(mode_path, f"{mode}.yaml")
            try:
                os.stat(mode_config)
            except FileNotFoundError:
                mode_config_template = os.path.join(pathlib.Path(__file__).parent.absolute(), f"templates/{mode}.yaml")
                shutil.copy2(mode_config_template, mode_config)



"""Test framework for testing physical platform integration."""

import asyncio
from datetime import datetime

from mpf.core.delays import DelayManager
from mpf.core.plugin import MpfPlugin
from mpf.core.utility_functions import Util


class MpfPlatformIntegrationTestRunner(MpfPlugin):

    """Runs a Platform Integration test provided by the command line.

    mpf -pit path_to/test_file.py
        -or-
    mpf -pit path.to.ModuleFile.ModuleClassName
    """

    __slots__ = ("delay", "_task", "_keep_alive", "_start_time", "_test_obj",
                 "trough_switches")

    def __init__(self, *args, **kwargs):
        """Initialize the test runner."""
        super().__init__(*args, **kwargs)
        self._keep_alive = None
        self._task = None
        self._test_obj = None
        self._start_time = None
        self.delay = None
        self.trough_switches = None

    @property
    def is_plugin_enabled(self):
        """Enable this plugin by the '-pit' command line arg."""
        return self.machine.options["platform_integration_test"]

    def initialize(self):
        """Initialize test runner and load test module."""
        self.configure_logging('MpfPlatformIntegration', 'basic', 'full')
        self.info_log("Initializing Platform Integration Test: Arg value is %s",
                      self.machine.options["platform_integration_test"])
        self.delay = DelayManager(self.machine)

        # Find the file with the test
        test_file = self.machine.options["platform_integration_test"]
        # If a path to a python file is found, convert it to a module path
        # This requires the import class to be the PascalCase version
        # of the python file's snake_case name.
        if test_file.endswith(".py"):
            parts = test_file.split("/")
            parts[-1] = parts[-1][:-3]
            parts.append(Util.snake_to_pascal(parts[-1]))
            test_file = ".".join(parts)

        self.info_log("Trying to import from module '%s'", test_file)
        # Look for a file path
        self._test_obj = Util.string_to_class(test_file)(self)

        # Wait until init phase 3 to setup test
        self.machine.events.add_handler("init_phase_3", self._setup_test)

    def keep_alive(self, value=True):
        """Configure whether the test runner will keep the MPF instance running after the test run finishes."""
        self._keep_alive = value

    def _setup_test(self, **kwargs):
        del kwargs
        self.machine.events.remove_handler(self._setup_test)

        # Create an event handler to begin the test run
        self.machine.events.add_handler(self._test_obj.test_start_event, self._run_test)

        self.trough_switches = []
        for ball_device in self.machine.ball_devices.items_tagged('trough'):
            jam_switch = ball_device.config.get('jam_switch')
            for s in Util.string_to_list(ball_device.config["ball_switches"]):
                if s != jam_switch:
                    self.trough_switches.append(s)

        # Pre-set initial switches defined in the test
        if self._test_obj.initial_switches is not None:
            for s in self._test_obj.initial_switches:
                # If a tuple is provided, the second value is the target state
                if isinstance(s, tuple):
                    s, s_state = s
                else:
                    s_state = 1
                if self.machine.switches[s].invert:
                    s_state ^= 1
                self.machine.switch_controller.process_switch(s, s_state)
            # Recheck all the ball counts
            for ball_device in self.machine.ball_devices.values():
                if hasattr(ball_device, "ball_count_handler"):
                    self.info_log("Resetting ball count for %s", ball_device)
                    ball_device.ball_count_handler.counter.trigger_recount()
                    # ball_device.ball_count_handler._count_valid.clear()
        # Pre-fill the trough with balls if no initial switches are defined
        else:
            for s in self.trough_switches:
                s_state = 0 if self.machine.switches[s].invert else 1
                self.info_log("Initializing trough switch %s", s)
                self.machine.switch_controller.process_switch(s, s_state)
            self.info_log("Re-initializing trough ball counts")
            for ball_device in self.machine.ball_devices.items_tagged('trough'):
                ball_device.ball_count_handler.counter.trigger_recount()
            self.machine.ball_controller.num_balls_known = len(self.trough_switches)

    def _run_test(self, **kwargs):
        del kwargs
        self.machine.events.remove_handler(self._run_test)
        self.log.info("Running platform integration test!")
        self._start_time = datetime.now()
        for playfield in self.machine.playfields.values():
            playfield.ball_search.disable()
        self._task = asyncio.create_task(self._test_obj.run_test())
        self._task.add_done_callback(self._stop_callback)

    def _stop_callback(self, future):
        Util.raise_exceptions(future)

        if self._task:
            self._task.cancel()
            self._task = None
        duration = datetime.now() - self._start_time
        msg = f"All tests completed successfully in {duration.seconds}.{duration.microseconds // 1000} seconds."
        if not self._keep_alive:
            self.machine.stop(msg)
        else:
            self.info_log("%s Keep Alive is enabled, runner will not exit game." % msg)

    def disable_ball_search(self):
        """Disable ball search."""
        for playfield in self.machine.playfields.values():
            playfield.ball_search.disable()

    def disable_ball_saves(self):
        """Disable all ball saves."""
        for ball_save in self.machine.ball_saves.values():
            ball_save.disable()

    def accelerate_high_score(self):
        """If there is a high score mode and the test achieves a high score, accelerate the text entry timeout."""
        if hasattr(self.machine.modes, "high_score"):
            self.machine.modes.high_score.high_score_config['enter_initials_timeout'] = 2

    def set_switch_sync(self, switch_name, state):
        """Set a switch to a given state, synchronously."""
        if state is None:
            state = int(not self.machine.switches[switch_name].state)
        if self.machine.switches[switch_name].invert:
            state ^= 1
        self.delay.remove(switch_name)
        self.info_log("Setting switch %s to state %s", switch_name, state)
        self.machine.switch_controller.process_switch(switch_name, state)

    async def wait_for_event(self, event, timeout=None, continue_on_timeout=False):
        """Wait for an event to be posted within a given timeout (secs)."""
        try:
            await asyncio.wait_for(self.machine.events.wait_for_event(event), timeout)
            self.info_log(f"Successfully caught event '{event}'")
            return True
        except asyncio.TimeoutError:
            if continue_on_timeout:
                return False
            self.machine.stop_with_exception({
                "exception": AssertionError(f"Awaited event '{event}' failed to trigger within {timeout}s.")
            })

    # pylint: disable-msg=too-many-arguments
    async def set_switch(self, switch_name, state=None, duration_secs=None, wait_after=None, blocking=True):
        """Set a switch to a given state.

        Parameters
        ----------
            - duration_secs: Number of seconds to hold the switch in that state. If none, will switch indefinitely.
            - wait_after: Number of seconds to wait after the switch before returning
            - blocking: If true, will wait for duration_secs before returning. If false, will fire-and-forget.

        """
        self.set_switch_sync(switch_name, state)
        if duration_secs:
            # Blocking mode: await the duration before returning
            if blocking:
                self.info_log(f" - waiting {duration_secs}s to toggle back switch {switch_name}")
                await asyncio.sleep(duration_secs)
                self.set_switch_sync(switch_name, int(not state))
            else:
                # Remove any existing reversion, add a new one if necessary
                self.delay.add(name=switch_name, ms=duration_secs * 1000,
                               callback=self.set_switch_sync,
                               switch_name=switch_name, state=int(not state))

        if wait_after:
            await asyncio.sleep(wait_after)

    async def start_game(self, num_players=1):
        """Start a game with the requested number of players."""
        self.info_log("Starting game.")
        await asyncio.sleep(1)
        start_button = self.machine.switches.items_tagged('start')[0]
        for _ in range(0, num_players):
            await self.set_switch(start_button.name, 1, 0.2)
            await asyncio.sleep(0.5)
        assert self.machine.game, "Game failed to start"

    async def eject_and_plunge_ball(self, plunger_switch_name, plunger_lane_settle_time=2, **kwargs):
        """Shuffle the trough switches and plunger to simulate an eject."""
        del kwargs
        self.info_log("Ejecting and plunging ball...")
        self.set_switch_sync(self.trough_switches[0], 0)
        await asyncio.sleep(0.03)
        self.set_switch_sync(self.trough_switches[-1], 0)
        await asyncio.sleep(0.1)
        self.set_switch_sync(self.trough_switches[0], 1)
        await asyncio.sleep(0.25)
        await self.set_switch(plunger_switch_name, 1, duration_secs=plunger_lane_settle_time)
        await asyncio.sleep(1)

    async def move_ball_from_drain_to_trough(self, **kwargs):
        """Move a ball from the drain device to the trough device."""
        del kwargs
        drain_switches = self.machine.ball_devices.items_tagged('drain')[0].config.get('ball_switches')
        self.set_switch_sync(drain_switches[-1], 0)
        await asyncio.sleep(0.25)
        self.set_switch_sync(self.trough_switches[-1], 1)
        await asyncio.sleep(0.25)



"""Contains the base classes for FAST expansion and breakout boards."""

import asyncio
from base64 import b16decode
from importlib import import_module

from packaging import version

from mpf.platforms.fast.fast_defines import (BREAKOUT_FEATURES,
                                             EXPANSION_BOARD_FEATURES)


class FastExpansionBoard:

    """A FAST Expansion board on the EXP connection."""

    # pylint: disable-msg=too-many-instance-attributes
    __slots__ = ["name", "communicator", "config", "platform", "log", "address", "model", "features", "breakouts",
                 "breakouts_with_leds", "firmware_version", "hw_verified", "led_fade_rate"]

    def __init__(self, name: str, communicator, address: str, config: dict) -> None:
        """Initializes a FAST Expansion Board.

        Parameters
        ----------
            name (str): Name of this board from the config file
            communicator (FastExpCommunicator): FAST EXP Communicator
            address (str): Two-character hex address of this board
            config (dict): This board's section from the config file

        Raises
        ------
            AssertionError: If a breakout board in the config is not valid.
        """
        self.name = name
        self.communicator = communicator
        self.config = config
        self.platform = communicator.platform
        self.log = communicator.log
        self.address = address
        self.model = config['model']

        self.led_fade_rate = None
        self.firmware_version = None
        self.hw_verified = False  # have we made contact with the board and verified it's the right hardware?

        self.log.debug('Creating FAST Expansion Board "%s" (%s Address: %s)',
                       self.name, self.model, self.address)

        self.features = EXPANSION_BOARD_FEATURES[self.model]  # ([local model numbers,], num of remotes) tuple
        self.breakouts = dict()
        self.breakouts_with_leds = list()

        if self.config['led_hz'] > 31.25:
            self.config['led_hz'] = 31.25

        # create the local breakouts
        for idx in range(len(self.features['local_breakouts'])):
            self.create_breakout({'port': str(idx), 'model': self.features['local_breakouts'][idx]})

        # create the remote breakouts
        for brk in self.config['breakouts']:
            if int(brk['port']) > self.features['breakout_ports']:
                # TODO change to config file error
                raise AssertionError(f'Breakout port {brk["port"]} is not available on {self}')

            self.create_breakout(brk)

    def create_breakout(self, config: dict) -> None:
        """Define a breakout board within an EXP board."""
        if BREAKOUT_FEATURES[config['model']].get('device_class'):
            module = import_module('mpf.platforms.fast.fast_exp_board')
            brk_board = module.FastBreakoutBoard(config, self)
        else:
            brk_board = FastBreakoutBoard(config, self)

        self.breakouts[config['port']] = brk_board
        self.platform.register_breakout_board(brk_board)

    def __repr__(self):
        """Return representation of this expansion board."""
        return f'EXP "{self.name}" ({self.model}, @{self.address})'

    def get_description_string(self) -> str:
        """Return description string."""
        # TODO add breakout boards
        return f"Expansion Board Model: {self.model_string},  Firmware: {self.firmware_version}"

    def verify_hardware(self, id_string: str, active_board: str) -> None:
        """Verifies an EXP or breakout board firmware versions.

        Parameters
        ----------
            id_string (str): ID string returned from the ID: FAST Serial Command
            active_board (str): 2 or 3 hex character address of the EXP or BRK board
        """
        self.log.info('Verifying hardware for %s with ID string "%s", board address %s',
                      self, id_string, active_board)

        exp_board = active_board[:2]
        brk_board = active_board[2:]  # will be empty if we got a 2-digit address for an EXP board

        try:
            _, product_id, firmware_version = id_string.split()
        except ValueError as e:
            if id_string == 'F':  # got an ID:F response which means this breakout is not actually there
                self.log.error('Breakout %s on %s is not responding', brk_board, self)
                raise AssertionError(f'Breakout {brk_board} on {self} is not responding') from e
            raise AssertionError(f'Invalid ID string {id_string} from {self}') from e

        assert exp_board == self.address
        self.firmware_version = firmware_version

        if brk_board:
            if version.parse(firmware_version) < version.parse(self.breakouts[brk_board].features['min_fw']):
                self.log.error('Firmware on breakout board %s is too old. Required: %s, Actual: %s. '
                               'Update at fastpinball.com/firmware',
                               product_id, self.breakouts[brk_board].features["min_fw"], firmware_version)
                self.platform.machine.stop(f'Firmware on breakout board {product_id} is too old. '
                                           f'Required: {self.breakouts[brk_board].features["min_fw"]}, '
                                           f'Actual: {firmware_version}. Update at fastpinball.com/firmware')

            brk = self.breakouts[brk_board]

            if product_id != brk.model:
                raise AssertionError(f'EXP Board {self}, Breakout Port {brk.index}: '
                                     f'Config is set for {brk.model}, but actually found "{id_string}"')
            brk.hw_verified = True

        else:
            if version.parse(firmware_version) < version.parse(self.features['min_fw']):
                self.log.error('Firmware on %s is too old. Required: %s, Actual: %s. '
                               'Update at fastpinball.com/firmware',
                               self, self.features["min_fw"], firmware_version)
                self.platform.machine.stop(f'Firmware on {self} is too old. Required: {self.features["min_fw"]}, '
                                           f'Actual: {firmware_version}. Update at fastpinball.com/firmware')

            if product_id != self.model:
                raise AssertionError(f"Expected {self.model} but got {id_string} from {self}")
            self.hw_verified = True

    async def reset(self):
        """Send a reset command to the EXP board."""
        await self.communicator.send_and_wait_for_response_processed(f'BR@{self.address}:', 'BR:P')

        # TODO move this to mixin classes for device types?
        if self.config['led_fade_time']:
            self.set_led_fade(self.config['led_fade_time'])

        # TODO re-initialize servos? Or call each breakout to do that?

    async def soft_reset(self):
        """Trigger a 'soft' reset of each breakout board on the EXP."""
        for breakout in self.breakouts.values():
            await breakout.soft_reset()

    def update_leds(self):
        """Look for LEDs that need updating and send to the platform.

        Called every tick to update the LEDs on this board.
        """
        for breakout_address in self.breakouts_with_leds:
            dirty_leds = {k: v.current_color for (k, v) in self.platform.fast_exp_leds.items()
                          if (v.dirty and v.address == breakout_address)}

            if dirty_leds:
                # TODO add the pre-encoded address to the defines file?
                # RD@<address>:, encode to binary then convert to hex chars
                msg_header = ''.join([f'{x:02X}' for x in f'RD@{breakout_address}:'.encode()])
                msg = f'{len(dirty_leds):02X}'

                for led_num, color in dirty_leds.items():
                    msg += f'{led_num[3:]}{color}'

                log_msg = f'RD@{breakout_address}:{msg}'  # pretty version of the message for the log

                try:
                    self.communicator.send_bytes(b16decode(f'{msg_header}{msg}'), log_msg)
                except Exception as e:
                    self.log.error(
                        f"Error decoding the following message for board {breakout_address} : {msg_header}{msg}")
                    self.log.debug("Attempted update that caused this error: %s", dirty_leds)
                    raise e

    def set_led_fade(self, rate: int) -> None:
        """Set LED fade rate in ms."""
        self.led_fade_rate = rate
        self.communicator.set_led_fade_rate(self.address, rate)


class FastBreakoutBoard:

    """A FAST Breakout board on the EXP connection."""

    __slots__ = ["config", "expansion_board", "log", "index", "platform", "communicator", "address", "features",
                 "leds", "led_fade_rate", "hw_verified", "model"]

    def __init__(self, config, expansion_board):
        """Initialize FastBreakoutBoard."""
        self.config = config
        self.expansion_board = expansion_board  # object
        self.log = expansion_board.log
        self.index = config['port']  # int, zero-based, 0-5
        self.log.debug("Creating FAST Breakout Board %s on %s", self.index, self.expansion_board)
        self.platform = expansion_board.platform
        self.communicator = expansion_board.communicator
        self.address = f'{self.expansion_board.address}{self.index}'  # string hex byte + nibble
        self.features = BREAKOUT_FEATURES[config['model']]
        self.leds = list()  # TODO move to mixin class
        self.led_fade_rate = 0
        self.hw_verified = False

        self.model = self.config['model']

        self.platform.machine.events.add_handler('init_phase_2', self._initialize)

    def __repr__(self):
        """Return representation of the breakout board."""
        return f"Breakout {self.model} on {self.expansion_board}"

    def _initialize(self, **kwargs):
        """Populate the LED objects."""
        del kwargs
        # TODO move to a mixin class based on device type
        found = False
        for number, led in self.platform.fast_exp_leds.items():
            if number.startswith(self.address):
                self.leds.append(led)
                found = True

        if found:
            self.expansion_board.breakouts_with_leds.append(self.address)

    async def soft_reset(self):
        """Reset the breakout board."""
        self.communicator.send_and_forget(f'RA@{self.address}:000000')
        await asyncio.sleep(.03)

        # Should we do something with servos? TODO
        # TODO move this to mixin classes for device types?



#!/usr/bin/python3
"""AFL fuzzer."""
import argparse
import asyncio
import gc
import os
import sys

import logging

import mpf.core
from mpf.core.logging import LogMixin
from mpf.core.utility_functions import Util
from mpf.tests.MpfTestCase import TestMachineController, UnitTestConfigLoader
from mpf.tests.loop import TimeTravelLoop, TestClock


class AflRunner(object):

    """AFL fuzzer."""

    __slots__ = ["loop", "clock", "machine", "debug", "machine_config_patches", "machine_config_defaults",
                "switch_list", "use_virtual", "_invalid_input", "_exception"]

    def __init__(self, use_virtual, debug):
        """Initialize fuzzer."""
        self.loop = None
        self.clock = None
        self.machine = None     # type: TestMachineController
        self.debug = debug
        self.machine_config_patches = dict()
        self.machine_config_patches['mpf'] = dict()
        self.machine_config_patches['mpf']['default_platform_hz'] = 1
        self.machine_config_patches['bcp'] = []
        self.machine_config_defaults = {}
        self.switch_list = []
        self.use_virtual = use_virtual
        self._invalid_input = False
        self._exception = None

    def _exception_handler(self, loop, context):
        try:
            loop.stop()
        except RuntimeError:
            pass

        self._exception = context

    def get_platform(self):
        """Return platform."""
        if self.use_virtual:
            return "virtual"

        return "smart_virtual"

    @staticmethod
    def get_config_file():
        """Return config file."""
        return "config.yaml"

    def get_options(self):
        """Return option arrays."""
        mpfconfig = os.path.abspath(os.path.join(
            mpf.core.__path__[0], os.pardir, 'mpfconfig.yaml'))

        return {
            'force_platform': self.get_platform(),
            'mpfconfigfile': mpfconfig,
            'configfile': Util.string_to_event_list(self.get_config_file()),
            'debug': False,
            'bcp': False,
            'no_load_cache': False,
            'platform_integration_test': False,
            'create_config_cache': True,
            'text_ui': False,
            'production': not self.debug,
        }

    def advance_time_and_run(self, delta=1.0):
        """Advance time and run."""
        try:
            self.loop.run_until_complete(asyncio.sleep(delay=delta, loop=self.loop))
            return
        except RuntimeError as e:
            if self._exception and "exception" in self._exception:
                raise self._exception['exception']
            elif self._exception:
                raise Exception(self._exception, e)
            raise e

    def setup(self, machine_path):
        """Set up fuzzer."""
        self.loop = TimeTravelLoop()
        self.loop.set_exception_handler(self._exception_handler)
        self.clock = TestClock(self.loop)

        config_loader = UnitTestConfigLoader(machine_path, ["config.yaml"], self.machine_config_defaults,
                                             self.machine_config_patches, {})

        config = config_loader.load_mpf_config()
        try:
            # remove virtual_platform_start_active_switches as it messes with out switch logic
            config.get_machine_config().pop("virtual_platform_start_active_switches")
        except KeyError:
            pass

        self.machine = TestMachineController(
            self.get_options(), config, self.machine_config_patches, self.machine_config_defaults, self.clock, dict(),
            True)

        try:
            self.loop.run_until_complete(self.machine.initialize())
        except RuntimeError as e:
            try:
                self.machine.stop()
            # pylint: disable-msg=broad-except
            except Exception:
                pass
            if self._exception and "exception" in self._exception:
                raise self._exception['exception']
            elif self._exception:
                raise Exception(self._exception, e)
            raise e

        self.machine.events.process_event_queue()
        self.advance_time_and_run(1)

        self.switch_list = sorted(self.machine.switches.keys())

        for switch in self.machine.switches:
            self.machine.switch_controller.process_switch_obj(switch, 0, True)

    def add_balls(self):
        """Add balls."""
        balls_to_add = 3
        for device in sorted(self.machine.ball_devices.values()):
            if "trough" in device.tags:
                for switch in device.ball_count_handler.counter.config.get('ball_switches', []):
                    self.machine.switch_controller.process_switch_obj(switch, 1, True)
                    balls_to_add -= 1
                    if balls_to_add == 0:
                        break
                if balls_to_add == 0:
                    break
                if device.ball_count_handler.counter.config.get('entrance_switch'):
                    self.machine.switch_controller.process_switch_obj(
                        device.ball_count_handler.counter.config['entrance_switch'], 1, True)
                    balls_to_add -= 1
                    if balls_to_add == 0:
                        break

        # let balls settle
        self.advance_time_and_run(10)

    def start_game(self):
        """Start game."""
        found_start_switch = False
        for switch in self.machine.switches:
            if "start" in switch.tags:
                found_start_switch = True
                self.machine.switch_controller.process_switch_obj(switch, 1, True)
                self.machine.switch_controller.process_switch_obj(switch, 0, True)

        if not found_start_switch:
            raise AssertionError("Did not find any switch tagged with tag \"start\".")

    def _abort(self, **kwargs):
        """Abort fuzzer run."""
        del kwargs
        self._invalid_input = True
        if not self.debug:
            os._exit(0)  # NOQA

    def run(self, actions, find_logic_bugs):
        """Run fuzzer."""
        if find_logic_bugs:
            self.machine.events.add_handler("balldevice_ball_missing", self._abort)
            self.machine.events.add_handler("found_new_ball", self._abort)
            self.machine.events.add_handler("mode_game_stopped", self._abort)

        for action in actions:
            if self._invalid_input:
                # bail out if we hit an invalid input. afl will notice this
                return

            if action & 0b10000000:
                if action == 0b11111111:
                    if self.machine.game and \
                            self.machine.game.balls_in_play < self.machine.ball_controller.num_balls_known:
                        self.machine.playfield.add_ball()
                        self.machine.game.balls_in_play += 1
                    else:
                        self._abort()
                else:
                    ms = int(action & 0b01111111)
                    ms *= ms
                    self.advance_time_and_run(ms / 1000.0)
            else:
                switch = int(action & 0b01111111)
                if switch >= len(self.switch_list):
                    self._abort()
                    return
                switch_obj = self.machine.switches[self.switch_list[switch]]
                state = switch_obj.hw_state ^ 1
                # print(switch_list[switch], state, switch_obj.hw_state)
                self.machine.switch_controller.process_switch_by_num(switch_obj.hw_switch.number, state,
                                                                     self.machine.default_platform)

        if find_logic_bugs:
            self.advance_time_and_run(60)
            if self._invalid_input:
                # might happen late
                return

            balls = 0
            for playfield in self.machine.playfields:
                balls += playfield.balls

            if balls != self.machine.game.balls_in_play:
                print("Balls in play:", self.machine.game.balls_in_play)
                print("Playfields:")
                for playfield in self.machine.playfields:
                    print(playfield.name, playfield.balls)
                print("Devices:")
                for device in self.machine.ball_devices:
                    print(device.name, device.balls, device.available_balls)
                raise AssertionError("Balls in play do not match balls on playfields.")

    def dump(self, wait, add_balls, start_game, actions):
        """Dump fuzzer input."""
        if add_balls:
            balls_to_add = 3
            for device in sorted(self.machine.ball_devices.values()):
                if "trough" in device.tags:
                    for switch in device.ball_count_handler.counter.config.get('ball_switches', []):
                        print("Enable switch {}".format(switch.name))
                        switch.hw_state = 1
                        balls_to_add -= 1
                        if balls_to_add == 0:
                            break

                    if balls_to_add == 0:
                        break
                    if device.ball_count_handler.counter.config.get('entrance_switch'):
                        print("Enable switch {}".format(device.config['entrance_switch'].name))
                        device.config['entrance_switch'].switch.hw_state = 1
                        balls_to_add -= 1
                        if balls_to_add == 0:
                            break
            print("Advance time 10s")

        if start_game:
            for switch in self.machine.switches:
                if "start" in switch.tags:
                    print("Enable/disable switch {}".format(switch.name))

        if wait > 0:
            print("Advance time {}s".format(wait))

        for action in actions:
            if action & 0b10000000:
                if action == 0b11111111:
                    print("Add ball into play")
                else:
                    ms = int(action & 0b01111111)
                    ms *= ms
                    print("Advance time by {} ms".format(ms))
            else:
                switch = int(action & 0b01111111)
                if switch >= len(self.switch_list):
                    print("Invalid switch. Abort!")
                    return
                switch_obj = self.machine.switches[self.switch_list[switch]]
                switch_obj.hw_state = state = switch_obj.hw_state ^ 1
                # print(switch_list[switch], state, switch_obj.hw_state)
                print("Toggle switch {}. New state: {}".format(switch_obj.name, state))


parser = argparse.ArgumentParser(
    description='Fuzz MPF using AFL')

parser.add_argument("-d",
                    action="store_true", dest="debug",
                    help="Turn on debug to reproduce fuzzer results")

parser.add_argument("-D",
                    action="store_true", dest="dump",
                    help="Dump test case.")

parser.add_argument("-u",
                    action="store_true", dest="unit_test",
                    help="Run in unit tests mode which fails early")

parser.add_argument("-b",
                    action="store_true", dest="add_balls",
                    help="Add balls to trough")

parser.add_argument("-w",
                    action="store", dest="wait", default=0,
                    help="Run machine for x seconds before forking")

parser.add_argument("-G",
                    action="store_true", dest="start_game",
                    help="Start game")

parser.add_argument("-L",
                    action="store_true", dest="find_logic_bugs",
                    help="Find game logic bugs only")

parser.add_argument("-v",
                    action="store_true", dest="use_virtual",
                    help="Use virtual instead of smart_virtual for low-level fuzzing")

parser.add_argument("machine_path", help="Path of the machine folder",
                    default=None, nargs='?')

args = parser.parse_args()

if args.debug:
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s : %(levelname)s : %(name)s : %(message)s')
else:
    logging.basicConfig(level=99)

if args.unit_test:
    LogMixin.unit_test = True

runner = AflRunner(use_virtual=args.use_virtual, debug=args.debug)
runner.setup(args.machine_path)

if args.dump:
    action_str = sys.stdin.buffer.read(-1)
    runner.dump(int(args.wait), args.add_balls, args.start_game, action_str)
    os._exit(-1)    # NOQA

if args.add_balls:
    runner.add_balls()

if args.start_game:
    if not args.add_balls:
        raise AssertionError("Cannot start game without balls. Use -b")
    runner.start_game()

if int(args.wait) > 0:
    runner.advance_time_and_run(int(args.wait))

if args.start_game and not runner.machine.game:
    raise AssertionError("Failed to start a game.")

# run GC once to clean everything up
gc.collect()

# keep effort minimal after those two lines. everything before this will execute only once.
# everything after this on every run
import afl  # NOQA
afl.init()

action_str = sys.stdin.buffer.read(-1)

runner.run(action_str, args.find_logic_bugs)

if args.debug:
    runner.advance_time_and_run(10)
else:
    os._exit(0)  # NOQA



"""Contains the DeviceManager base class."""
import asyncio

from typing import Callable, Tuple, List, Generator, Dict

from mpf.core.utility_functions import Util
from mpf.core.mpf_controller import MpfController

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.device import Device  # pylint: disable-msg=cyclic-import,unused-import


class DeviceManager(MpfController):

    """Manages all the devices in MPF."""

    config_name = "device_manager"

    __slots__ = ["_monitorable_devices", "collections", "_device_classes"]

    def __init__(self, machine):
        """Initialize device manager."""
        super().__init__(machine)

        self._monitorable_devices = {}

        self.collections = dict()
        self._device_classes = {}           # type: Dict[str, Device]

        # this has to happen after mode load
        self.machine.events.add_async_handler('init_phase_1',
                                              self._load_device_modules, priority=5)

        self.machine.events.add_handler('init_phase_2',
                                        self.create_machinewide_device_control_events,
                                        priority=2)

    def get_monitorable_devices(self):
        """Return all devices which are registered as monitorable."""
        return self._monitorable_devices

    def register_monitorable_device(self, device):
        """Register a monitorable device.

        Args:
        ----
            device: The device to register.
        """
        if device.collection not in self._monitorable_devices:
            self._monitorable_devices[device.collection] = {}
        self._monitorable_devices[device.collection][device.name] = device

    def notify_device_changes(self, device, notify, old, value):
        """Notify subscribers about changes in a registered device.

        Args:
        ----
            device: The device that changed.
            notify: Attribute name which changed.
            old: The old value.
            value: The new value.

        """
        self.machine.bcp.interface.notify_device_changes(device, notify, old, value)

    async def _load_device_modules(self, **kwargs):
        del kwargs
        # step 1: create devices in machine collection
        self.debug_log("Creating devices...")
        for collection_name in self.machine.config['mpf']['device_modules'].keys():
            # create the collection
            collection = DeviceCollection(self.machine, collection_name, collection_name)

            self.collections[collection_name] = collection
            setattr(self.machine, collection_name, collection)

            # Get the config section for these devices
            config = self.machine.config.get(collection_name, None)

            # create the devices
            if config:
                self.create_devices(collection_name, config)

        self.machine.mode_controller.create_mode_devices()

        # step 2: load config and validate devices
        self.load_devices_config(validate=True)
        await self.machine.mode_controller.load_mode_devices()

        # step 3: initialize devices (mode devices will be initialized when mode is started)
        await self.initialize_devices()

    def stop_devices(self):
        """Stop all devices in the machine."""
        for collection in self.collections.values():
            for device in collection.values():
                if hasattr(device, "stop_device"):
                    device.stop_device()

    def create_devices(self, collection_name, config):
        """Create devices for a collection."""
        if collection_name not in self._device_classes:
            cls = self._device_classes[collection_name] = Util.string_to_class(
                self.machine.config['mpf']['device_modules'][collection_name])
        else:
            cls = self._device_classes[collection_name]

        collection = getattr(self.machine, collection_name)

        # if this device class has a device_class_init classmethod, run it now
        if hasattr(cls, 'device_class_init'):
            # don't want to use try here in case the called method has an error
            cls.device_class_init(self.machine)

        # create the devices
        for device_name in config:
            if not config[device_name] and not cls.allow_empty_configs:
                self.raise_config_error("Device {}:'{}' has an empty config.".format(collection_name, device_name), 2,
                                        context=collection_name + "." + device_name)

            elif not isinstance(config[device_name], dict):
                self.raise_config_error("Device {}:'{}' does not have a valid config. Expected a dictionary.".format(
                    collection_name, device_name), 3, context=collection_name + "." + device_name)

            collection[device_name] = cls(self.machine, device_name)

    def load_devices_config(self, validate=True):
        """Load all devices."""
        if validate:
            for collection_name in self.machine.config['mpf']['device_modules'].keys():
                if collection_name not in self.machine.config:
                    continue

                # Get the config section for these devices
                collection = getattr(self.machine, collection_name)
                config = self.machine.config[collection_name]
                if not config:
                    self.machine.config[collection_name] = config = {}
                if not isinstance(config, dict):
                    self.raise_config_error("Format of collection {} is invalid.".format(collection_name), 1)

                # validate config
                for device_name in config:
                    config[device_name] = collection[device_name].prepare_config(config[device_name], False)
                    config[device_name] = collection[device_name].validate_and_parse_config(config[device_name], False)

        for collection_name in self.machine.config['mpf']['device_modules'].keys():
            if collection_name not in self.machine.config:
                continue

            # Get the config section for these devices
            collection = getattr(self.machine, collection_name)
            config = self.machine.config[collection_name]

            # load config
            for device_name in config:
                collection[device_name].load_config(config[device_name])

    async def initialize_devices(self):
        """Initialize devices."""
        futures = []
        for collection_name in self.machine.config['mpf']['device_modules'].keys():
            if collection_name not in self.machine.config:
                continue

            # Get the config section for these devices
            collection = getattr(self.machine, collection_name)
            config = self.machine.config[collection_name]

            # add machine wide
            for device_name in config:
                futures.append(collection[device_name].device_added_system_wide())

        await asyncio.wait([asyncio.create_task(futures) for futures in futures])

    # pylint: disable-msg=too-many-nested-blocks
    def get_device_control_events(self, config) -> Generator[Tuple[str, Callable, int, "Device"], None, None]:
        """Scan a config dictionary for control_events.

         Yields events, methods, delays, and devices for all the devices and
         control_events in that config.

        Args:
        ----
            config: An MPF config dictionary (either machine-wide or mode-
                specific).

        Returns a generator of 4-item tuples
        ------------------------------------
            * The event name
            * The callback method of the device
            * The delay in ms
            * The device object
        """
        config_spec = self.machine.config_validator.get_config_spec()
        for collection_name, collection in self.collections.items():
            config_section = collection.config_section
            if config_section in config:
                for device, settings in iter(config[config_section].items()):

                    control_events = [x for x in settings if
                                      x.endswith('_events') and x != "control_events"]
                    device_obj = collection[device]

                    for control_event in control_events:
                        # get events from this device's config
                        if not isinstance(settings[control_event], dict):
                            if config_spec[config_section][control_event] == "ignore":
                                continue
                            raise AssertionError(
                                "Type of {}:{} should be event_handler|event_handler:ms| in config_spec".format(
                                    collection_name, control_event))
                        # try the new style first
                        try:
                            method = getattr(device_obj, "event_{}".format(control_event[:-7]))
                        except AttributeError:
                            raise AssertionError("Class {} needs to have method event_{} to handle {}".format(
                                device_obj, control_event[:-7], control_event
                            ))
                        for event, delay in settings[control_event].items():
                            yield (event,
                                   method,
                                   delay,
                                   device_obj)

    def create_machinewide_device_control_events(self, **kwargs):
        """Create machine wide control events."""
        del kwargs
        for event, method, delay, _ in (
                self.get_device_control_events(self.machine.config)):

            if delay:
                self.machine.events.add_handler(
                    event=event,
                    handler=self._control_event_handler,
                    callback=method,
                    ms_delay=delay,
                    delay_mgr=self.machine.delay)
            else:
                self.machine.events.add_handler(
                    event=event,
                    handler=method)

    def _control_event_handler(self, callback, ms_delay, delay_mgr=None, **kwargs):
        del kwargs

        self.debug_log("_control_event_handler: callback: %s,", callback)
        delay_mgr.add(ms=ms_delay, callback=callback)


class DeviceCollection(dict):

    """A collection of Devices.

    One instance of this class will be created for each different type of
    hardware device (such as coils, lights, switches, ball devices, etc.).
    """

    __slots__ = ["machine", "name", "config_section", "_tag_cache"]

    def __init__(self, machine, collection, config_section):
        """Initialize device collection."""
        super().__init__()

        self.machine = machine
        self.name = collection
        self.config_section = config_section
        self._tag_cache = dict()

    def __hash__(self):
        """Hash collection."""
        return hash((self.name, self.machine))

    def __delitem__(self, key):
        """Delete item for key."""
        # clear the tag cache
        self._tag_cache = dict()
        return super().__delitem__(key)

    def __getattr__(self, attr):
        """Return device by key.

        This method is DEPRECATED and will be removed soon.
        """
        # We used this to allow the programmer to access a hardware item like
        # self.coils.coilname
        self.machine.log.warning("Accessing devices by attribute is deprecated and will be removed. "
                                 "Access by value instead: %s.%s -> %s['%s']",
                                 self.name, attr, self.name, attr)
        try:
            return self[attr]
        except KeyError:
            raise KeyError('Error: No device exists with the name:', attr)

    def __iter__(self):
        """Iterate collection.

        This method is DEPRECATED and will be removed soon. Use .values() instead.
        """
        self.machine.log.warning("Iterating device collections directly is deprecated and will be removed. "
                                 "Access by value(): device_collections[%s] -> device_collections['%s'].values()",
                                 self.name, self.name)
        for item in self.values():
            yield item

    def items_tagged(self, tag) -> List["Device"]:
        """Return of list of device objects which have a certain tag.

        Args:
        ----
            tag: A string of the tag name which specifies what devices are
                returned.  A value of "*" returns all devices.

        Returns a list of device objects. If no devices are found with that tag, it
        will return an empty list.
        """
        if tag == "*":
            return self.values()

        items_in_tag_cache = self._tag_cache.get(tag, None)
        if items_in_tag_cache is not None:
            return items_in_tag_cache

        items = [item for item in self.values() if hasattr(item, "tags") and tag in item.tags]
        self._tag_cache[tag] = items
        return items



"""A switch input on a PKONE Extension board."""
import logging
from collections import namedtuple

from mpf.core.platform import SwitchConfig
from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.platforms.pkone.pkone import PKONEHardwarePlatform    # pylint: disable-msg=cyclic-import,unused-import

PKONESwitchNumber = namedtuple("PKONESwitchNumber", ["board_address_id", "switch_number"])


# pylint: disable-msg=too-few-public-methods
class PKONESwitch(SwitchPlatformInterface):

    """An PKONE input on a PKONE Extension board."""

    __slots__ = ["log"]

    def __init__(self, config: SwitchConfig, number: PKONESwitchNumber, platform: "PKONEHardwarePlatform") -> None:
        """Initialize switch."""
        super().__init__(config, number, platform)
        self.log = logging.getLogger('PKONESwitch')

    def get_board_name(self):
        """Return PKONE Extension addr."""
        if self.number.board_address_id not in self.platform.pkone_extensions.keys():
            return "PKONE Unknown Board"
        return "PKONE Extension Board {}".format(self.number.board_address_id)



"""FAST Emulator Serial Communicator."""
# mpf/platforms/fast/communicators/emu.py

from packaging import version

from mpf.platforms.fast.communicators.base import FastSerialCommunicator

MIN_FW = version.parse('0.00')  # override in subclass


class FastEmuCommunicator(FastSerialCommunicator):

    """Handles the serial communication to the EMU processor on a FAST Retro Controller."""

    IGNORED_MESSAGES = []

    async def init(self):
        """Initialize the Emulation communicator. Not currently implemented."""

    async def soft_reset(self):
        """Soft reset the Emulation communicator. Not currently implemented."""


"""Test the bonus mode."""
from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import MpfTestCase
from mpf._version import version, extended_version


class TestMachineVariables(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/machine_vars/'

    def _get_mock_data(self):
        return {"machine_vars": {"player2_score": {"value": 118208660, 'expire': None, 'expire_secs': None},
                                 "player3_score": {"value": 17789290, 'expire': None, 'expire_secs': None},
                                 "player4_score": {"value": 3006600, 'expire': None, 'expire_secs': None},
                                 "another_score": {"value": 123, 'expire': None, 'expire_secs': None},
                                 "expired_value": {"value": 23,
                                                   "expire": self.clock.get_datetime().timestamp() - 100,
                                                   'expire_secs': 3600},
                                 "not_expired_value": {"value": 24,
                                                       "expire": self.clock.get_datetime().timestamp() + 100,
                                                       'expire_secs': 3600},
                                 "test1": {"value": 42, 'expire': None, 'expire_secs': None}},
                }

    def testSystemInfoVariables(self):
        self.assertTrue(self.machine.variables.is_machine_var("mpf_version"))
        self.assertTrue(self.machine.variables.is_machine_var("mpf_extended_version"))
        self.assertTrue(self.machine.variables.is_machine_var("python_version"))
        self.assertTrue(self.machine.variables.is_machine_var("platform"))
        self.assertTrue(self.machine.variables.is_machine_var("platform_system"))
        self.assertTrue(self.machine.variables.is_machine_var("platform_release"))
        self.assertTrue(self.machine.variables.is_machine_var("platform_version"))
        self.assertTrue(self.machine.variables.is_machine_var("platform_machine"))
        self.assertEqual(version, self.machine.variables.get_machine_var("mpf_version"))
        self.assertEqual(extended_version, self.machine.variables.get_machine_var("mpf_extended_version"))
        self.assertEqual(0.5, self.machine.variables.get_machine_var("master_volume"))

    def testTime(self):
        current_date = self.machine.clock.get_datetime()
        self.assertPlaceholderEvaluates(current_date.second, "machine.time.second")
        self.assertPlaceholderEvaluates(current_date.minute, "machine.time.minute")
        self.assertPlaceholderEvaluates(current_date.hour, "machine.time.hour")
        self.assertPlaceholderEvaluates(current_date.day, "machine.time.day")
        self.assertPlaceholderEvaluates(current_date.month, "machine.time.month")
        self.assertPlaceholderEvaluates(current_date.year, "machine.time.year")

        placeholder = self.machine.placeholder_manager.build_int_template("machine.time.second")
        value, future = placeholder.evaluate_and_subscribe({})
        self.assertEqual(current_date.second, value)
        self.assertFalse(future.done())

        self.advance_time_and_run(1.1)
        self.assertTrue(future.done())

        self.mock_event("test_event3")
        self.mock_event("test_event4")
        self.advance_time_and_run(60)
        self.assertEventCalled("test_event3", 1)
        self.assertEventCalled("test_event4", 1)
        self.advance_time_and_run(60)
        self.assertEventCalled("test_event3", 2)
        self.assertEventCalled("test_event4", 2)

    def testVarLoadAndRemove(self):
        self.assertFalse(self.machine.variables.is_machine_var("expired_value"))
        self.assertTrue(self.machine.variables.is_machine_var("not_expired_value"))
        self.assertTrue(self.machine.variables.is_machine_var("player2_score"))
        # previously-persisted variables should continue to persist
        self.assertTrue(self.machine.variables.machine_vars["player2_score"]["persist"])
        # random variable does not persist
        self.machine.variables.set_machine_var("temporary_variable", 1000)
        self.assertEqual(1000, self.machine.variables.get_machine_var("temporary_variable"))
        self.assertFalse(self.machine.variables.machine_vars["temporary_variable"]["persist"])
        # configured to persist
        self.assertTrue(self.machine.variables.machine_vars["test1"]["persist"])
        self.assertTrue(self.machine.variables.machine_vars["test2"]["persist"])
        # configured to not persist
        self.assertFalse(self.machine.variables.machine_vars["test3"]["persist"])
        self.assertEqual(118208660, self.machine.variables.get_machine_var("player2_score"))

        self.machine.variables.remove_machine_var("player2_score")

        self.assertFalse(self.machine.variables.is_machine_var("player2_score"))
        self.assertTrue(self.machine.variables.is_machine_var("player3_score"))

        self.machine.variables.machine_var_data_manager._trigger_save = MagicMock()
        self.machine.variables.remove_machine_var_search(startswith="player", endswith="_score")
        self.assertFalse(self.machine.variables.is_machine_var("player2_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player3_score"))

        self.assertEqual(123, self.machine.variables.get_machine_var("another_score"))

        self.machine.variables.configure_machine_var("test3", expire_secs=100, persist=True)
        self.machine.variables.set_machine_var("test3", "Hello")
        ts = self.machine.clock.get_datetime().timestamp()

        self.advance_time_and_run(10)

        self.machine.variables.machine_var_data_manager._trigger_save.assert_called_with()
        self.assertEqual({
                          'another_score': {'value': 123, 'expire': None, 'expire_secs': None},
                          "not_expired_value": {'value': 24, 'expire': None, 'expire_secs': None},
                          'master_volume': {'value': 0.5, 'expire': None, 'expire_secs': None},
                          'test1': {'value': 42, 'expire': None, 'expire_secs': None},
                          'test2': {'value': '5', 'expire': None, 'expire_secs': None},
                          'test3': {'expire': ts + 100, 'expire_secs': 100, 'value': 'Hello'}},
                         self.machine.variables.machine_var_data_manager.data)

    def testVarSetAndGet(self):
        self.assertEqual(118208660, self.machine.variables["player2_score"])

        self.assertEqual(None, self.machine.variables["player1_score"])
        self.machine.variables["player1_score"] = 123
        self.assertEqual(123, self.machine.variables["player1_score"])

        self.assertEqual(17789290, self.machine.variables.get("player3_score"))


class TestMalformedMachineVariables(MpfTestCase):

    def _get_mock_data(self):
        return {"machine_vars": {"player2_score": {"value": 118208660, 'expire': None, 'expire_secs': None},
                                 "player3_score": {"value": 17789290, 'expire': None, 'expire_secs': None},
                                 "player4_score": {"value": 3006600},
                                 "player5_score": 123,
                                 "player6_score": {"asd": 3006600, 'expire': None, 'expire_secs': None},
                                 "value": 0}}

    def testVarLoads(self):
        self.assertTrue(self.machine.variables.is_machine_var("player2_score"))
        self.assertEqual(118208660, self.machine.variables.get_machine_var("player2_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player5_score"))
        self.assertEqual(None, self.machine.variables.get_machine_var("player5_score"))

